"""
Utilities for creating/reading/modifying archives
"""

from __future__ import annotations

import asyncio
import enum
import functools
import io
import os
import tarfile
import zipfile
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path, PurePosixPath
from typing import (IO, Awaitable, Callable, Iterable, NamedTuple, Sequence, TypeVar, Union, cast)
from typing_extensions import Literal

from .. import fs
from ..event import CancellationToken, Event, raise_if_cancelled
from ..fs import FilePredicate, IfDirectoryExists, IfExists, Pathish

_AR_OPS_POOL = ThreadPoolExecutor(4)

T = TypeVar('T')


class ExtractInfo(NamedTuple):
    """
    Information about the extraction of a file from an archive.
    """
    #: Path within the archive from where the item was extracted
    ar_path: PurePosixPath
    #: Path on the system to where the item was extracted
    dest_path: Path
    #: The index of the file in the extraction process.
    n: int
    #: The total number of files that will be extracted.
    total_count: int


class AddInfo(NamedTuple):
    """
    Information about the addition of a file into an archive.
    """
    ar_path: PurePosixPath
    "Path within the archive that the file is being written to"


ExtractFileEvent = Event[ExtractInfo]
AddFileEvent = Event[AddInfo]


def ar_thread_pool() -> ThreadPoolExecutor:
    """
    The thread pool used to execute archive operations
    """
    return _AR_OPS_POOL


def _run_ar_op(fn: Callable[[], T]) -> Awaitable[T]:
    loop = asyncio.get_event_loop()
    return loop.run_in_executor(ar_thread_pool(), fn)


Format = Literal['zip', 'tgz', 'tbz', 'txz']
FormatOrAuto = Literal[Format, 'auto']


def is_tar_format(fmt: Format) -> bool:
    """Return ``True`` iff this archive is a tarfile format archive."""
    return fmt in ('tgz', 'tbz', 'txz')


class _Mode(enum.Enum):
    """
    The open-mode for an archive.
    """
    #: Write a new archive file (truncates a previous archive)
    WriteNew = 'write-new'
    #: Read an existing archive
    Read = 'read'
    #: Append to an existing archive
    Append = 'append'


def _open_zip(path: Path, mode: _Mode) -> zipfile.ZipFile:
    """Return an opened ZipFile"""
    mode_str = {
        _Mode.Append: 'a',
        _Mode.WriteNew: 'w',
        _Mode.Read: 'r',
    }[mode]
    mode_str = cast(Literal['a'], mode_str)
    return zipfile.ZipFile(str(path), mode_str, zipfile.ZIP_DEFLATED)


def _open_tar(path: Path, mode: _Mode, format_: Format) -> tarfile.TarFile:
    """Return an opened TarFile with the appropriate compression"""
    algo_str = {
        'tgz': 'gz',
        'tbz': 'bz2',
        'txz': 'xz',
    }[format_]
    mode_str = {
        _Mode.Read: 'r:' + algo_str,
        _Mode.Append: 'a:',
        _Mode.WriteNew: 'w:' + algo_str,
    }[mode]
    return tarfile.open(str(path), mode_str)


class MemberFromFile(NamedTuple):
    """
    Define an archive member by copying a file from the system into the archive.
    """
    #: The path on the system to copy from
    path: Pathish
    #: The destination for the member within the archive structure
    ar_path: Pathish


class MemberFromBytes(NamedTuple):
    """
    Define an archive member by writing a buffer or string into the archive.
    """
    #: The path within the archive to store to
    ar_path: Pathish
    #: The content of the archive member. Should be `bytes` or
    #: `str`. `str` will be encoded as UTF-8.
    content: str | bytes


NewMember = Union[MemberFromFile, MemberFromBytes]


def _add_to_tar(tf: tarfile.TarFile, item: NewMember, on_add: AddFileEvent | None) -> None:
    """Add the given item to a tar archive file"""
    if on_add:
        on_add.emit(AddInfo(PurePosixPath(item.ar_path)))
    if isinstance(item, MemberFromFile):
        tf.add(str(item.path), str(item.ar_path) if item.ar_path else None)
    else:
        assert isinstance(item, MemberFromBytes), f'Invalid item for tar archive {item!r}'
        info = tarfile.TarInfo(str(item.ar_path))
        buf = item.content
        if isinstance(buf, str):
            buf = buf.encode('utf-8')
        info.size = len(buf)
        bio = io.BytesIO(buf)
        tf.addfile(info, bio)


def _create_tar(path: Path, items: Iterable[NewMember], format: Format, on_add: AddFileEvent | None,
                cancel: CancellationToken | None) -> Path:
    """Create a tar archive for the given items."""
    with _open_tar(path, _Mode.WriteNew, format) as tfd:
        for item in items:
            raise_if_cancelled(cancel)
            _add_to_tar(tfd, item, on_add)
    return path


def _add_to_zip(zfd: zipfile.ZipFile, item: NewMember, on_add: AddFileEvent | None) -> None:
    """Add a single item to a Zip archive"""
    if on_add:
        on_add.emit(AddInfo(PurePosixPath(item.ar_path)))
    if isinstance(item, MemberFromFile):
        zfd.write(str(item.path), str(item.ar_path) if item.ar_path else None)
    else:
        assert isinstance(item, MemberFromBytes), f'Invalid item for zip archive {item!r}'
        buf = item.content
        zfd.writestr(str(item.ar_path), buf)


def _create_zip(path: Path, items: Iterable[NewMember], on_add: AddFileEvent | None,
                cancel: CancellationToken | None) -> Path:
    """Create a zip archive for the given items"""
    with _open_zip(path, _Mode.WriteNew) as zfd:
        for item in items:
            raise_if_cancelled(cancel)
            _add_to_zip(zfd, item, on_add)
    return path


def _get_format(path: Path, format_: FormatOrAuto) -> Format:
    """Get the format to use with ``path``"""
    if format_ != 'auto':
        return format_
    if path.name.endswith('.zip'):
        return 'zip'
    if path.name.endswith('.tar.gz') or path.suffix in ('.tgz', '.tar'):
        return 'tgz'
    if path.name.endswith('.tar.bz2') or path.suffix in ('.tbz', '.tbz2'):
        return 'tbz'
    if path.name.endswith('.tar.xz') or path.suffix in ('.txz'):
        return 'txz'
    raise RuntimeError(f'Unable to guess archive format from the filename ({path.name})')


async def create(path: Pathish,
                 items: Iterable[NewMember],
                 *,
                 format: FormatOrAuto = 'auto',
                 on_add: AddFileEvent | None = None,
                 cancel: CancellationToken | None = None) -> Path:
    """
    Create an archive.

    :param path: The full file path for the destination of the archive.
    :param items: The items that should be put in the archive. Refer to
        :class:`MemberFromFile` and :class:`MemberFromBytes` to understand this
        parameter.
    :param format: The format of archive to create. If :attr:`.Guess`, the type
        of archive will be guessed from the file extension of ``path``.
    :param cancel: A cancellation token for the archive creation.

    :returns: The path to the generated archive.
    """
    path = Path(path)
    format = _get_format(path, format)

    if is_tar_format(format):
        return await _run_ar_op(lambda: _create_tar(Path(path), items, cast(Format, format), on_add, cancel))
    if format == 'zip':
        return await _run_ar_op(lambda: _create_zip(Path(path), items, on_add, cancel))
    raise RuntimeError(f'Invalid `format` given ({repr(format)})')


def dir_items(root: Path,
              *,
              only_if: FilePredicate | None = None,
              unless: FilePredicate | None = None,
              prefix: Pathish | None = None) -> Iterable[MemberFromFile]:
    """
    Return a single-pass iterable of :class:`MemberFromFile` objects.

    :param root: The directory to enumerate. Archive member paths will be
        relative to this directory.
    :param only_if: Predicate to filter files to add.
    :param unless: Predicate to filter files to exclude.
    :param prefix: If provided, this directory prefix will be prepended to
        every item's path within the archive.

    .. note:: ``only_if`` is evaluated before ``unless``.
    """
    for item in root.glob('**/*'):
        if only_if and not only_if(item):
            continue
        if unless and unless(item):
            continue
        relpath = PurePosixPath(item.relative_to(root).as_posix())
        ar_path = relpath
        if prefix is not None:
            ar_path = PurePosixPath(prefix) / ar_path
        yield MemberFromFile(path=item, ar_path=ar_path)


async def create_from_dir(dirpath: Pathish,
                          *,
                          destination: Pathish | None = None,
                          only_if: FilePredicate | None = None,
                          unless: FilePredicate | None = None,
                          prefix: PurePosixPath | str | None = None,
                          format: FormatOrAuto = 'auto',
                          cancel: CancellationToken | None = None) -> Path:
    """
    Create an archive from the contents of a directory.

    :param dirpath: The directory to archive. The contents of this directory
        will be added to the archive, and the paths within the archive will
        be relative to this directory.
    :param destination: The destination of the archive. If not given, will
        place the archive as a sibling of ``dirpath`` with a file extension
        appended to the filename.
    :param only_if: A predicate of which files to add.
    :param unless: A predicate of which files to exclude.
    :param prefix: If given, this will be created as a top-level directory to
        prefix all files in the archive.
    :param format: The format to use. If not given, deduced from
        ``destination``. If neither is given, will create a ``.tgz`` archive.
    :param cancel: A cancellation token to interrupt the archive creation.
    """
    dirpath = Path(dirpath).absolute()
    if destination is None:
        default_ext = {
            'auto': '.tgz',
            'tgz': '.tgz',
            'txz': '.txz',
            'tbz': '.tbz',
            'zip': '.zip',
        }[format]
        destination = dirpath.parent / (dirpath.name + default_ext)
    destination = destination or (dirpath.parent / (dirpath.name + '.tgz'))
    return await create(destination,
                        dir_items(
                            dirpath,
                            only_if=only_if,
                            unless=unless,
                            prefix=PurePosixPath(prefix) if prefix else None,
                        ),
                        format=format,
                        cancel=cancel)


async def _prep_expand(dest: Path, if_exists: IfDirectoryExists, cancel: CancellationToken | None) -> bool:
    """Prepare a destination for archive expansion"""
    if not dest.exists():
        # Destination isn't already there. Create the directory in preparation for expansion
        dest.mkdir(parents=True, exist_ok=True)
        return True
    # Decide what to do next, since the destination directory exists
    if if_exists == 'replace':
        # We're replacing the directory, so remove the existing contents completely
        await fs.remove(dest, recurse=True, cancel=cancel)
        return True
    if if_exists == 'keep':
        # We're going to keep the destination and not perform archive expansion
        return False
    if if_exists == 'merge':
        # We're merging over the top, so keep the top-level directory as-is
        return True
    if if_exists == 'fail':
        # We don't expect it to already exist, so raise an error
        raise FileExistsError(f'Archive expansion destination ({dest}) already exists')
    # Unreachable:
    assert False, f'Invalid value for if_exists: {repr(if_exists)}'


def _write_file(fd: IO[bytes], dest: Path) -> None:
    """
    Write the bytes of an IO-like object to the given path. Used by the
    archive expansion to write an archive member to the filesystem.
    """
    dest.parent.mkdir(exist_ok=True, parents=True)
    with dest.open('wb') as out:
        while True:
            buf = fd.read(1024 * 1)
            if not buf:
                break
            out.write(buf)


MemInfo = Union[tarfile.TarInfo, zipfile.ZipInfo]


def _path_of(m: MemInfo) -> Path:
    if isinstance(m, tarfile.TarInfo):
        return Path(m.name)
    return Path(m.filename)


def _is_dir(m: MemInfo) -> bool:
    if isinstance(m, tarfile.TarInfo):
        return m.isdir()
    return m.is_dir()


def _is_file(m: MemInfo) -> bool:
    if isinstance(m, tarfile.TarInfo):
        return m.isfile()
    return not m.is_dir()


def _open(ar: zipfile.ZipFile | tarfile.TarFile, m: MemInfo) -> IO[bytes]:
    if isinstance(ar, tarfile.TarFile):
        assert isinstance(m, tarfile.TarInfo)
        mem = ar.extractfile(m)
        assert mem
        return mem
    assert isinstance(m, zipfile.ZipInfo)
    return ar.open(m)


async def _expand_archive(
    archive: zipfile.ZipFile | tarfile.TarFile,
    dest: Path,
    only_if: FilePredicate,
    unless: FilePredicate,
    strip_components: int,
    if_file_exists: IfExists,
    cancel: CancellationToken | None,
    on_extract: ExtractFileEvent | None,
) -> Path:
    # Switch based on archive type
    is_tar = isinstance(archive, tarfile.TarFile)

    all_members = cast(Sequence[Union[zipfile.ZipInfo, tarfile.TarInfo]],
                       archive.getmembers() if is_tar else archive.infolist())  # type: ignore
    filtered = (mem for mem in all_members if (only_if(_path_of(mem)) and not unless(_path_of(mem))))

    dest_map: list[tuple[Union[zipfile.ZipInfo, tarfile.TarInfo], Path]] = []

    for member in filtered:
        path_elems = PurePosixPath(_path_of(member)).parts
        new_elems = path_elems[min(strip_components, len(path_elems)):]
        if not len(new_elems):
            continue
        new_dest = functools.reduce(lambda a, b: a / b, new_elems, dest)
        if new_dest.exists():
            if if_file_exists == 'fail':
                raise FileExistsError(f'Destination file already exists: {new_dest}')
            if if_file_exists == 'keep':
                # Skip this archive member
                continue
            if if_file_exists == 'replace':
                if _is_dir(member):
                    # We keep around directories, since we want to merge over the top of them
                    pass
                else:
                    os.unlink(new_dest)
        dest_map.append((member, new_dest))

    dirs: list[tuple[tarfile.TarInfo, Path]] = []

    for idx, pair in enumerate(dest_map):
        raise_if_cancelled(cancel)
        mem, dest_filepath = pair
        if on_extract:
            info = ExtractInfo(PurePosixPath(_path_of(mem)), dest_filepath, idx, len(dest_map))
            on_extract.emit(info)
        if _is_file(mem):
            fd = _open(archive, mem)
            assert fd
            await _run_ar_op(lambda: _write_file(fd, dest_filepath))  # pylint: disable=cell-var-from-loop
        if _is_dir(mem) and is_tar:
            assert isinstance(mem, tarfile.TarInfo)
            dirs.append((mem, Path(dest_filepath)))

    dirs.sort(key=lambda pair: pair[1])
    dirs.reverse()
    for mem, dirpath in dirs:
        assert is_tar, 'Code path should not be taken unless expanding a Tar archive'
        os.chmod(dirpath, mem.mode)
        os.utime(dirpath, (mem.mtime, mem.mtime))

    return dest


async def expand(path: Pathish,
                 *,
                 destination: Pathish,
                 only_if: FilePredicate | None = None,
                 unless: FilePredicate | None = None,
                 strip_components: int = 0,
                 format: FormatOrAuto = 'auto',
                 if_exists: IfDirectoryExists = 'fail',
                 if_file_exists: IfExists = 'fail',
                 cancel: CancellationToken | None = None,
                 on_extract: ExtractFileEvent | None = None) -> Path:
    """
    Expand the contents of an archive into a directory.

    :param path: The path to an archive.
    :param destination: The directory to expand *into*. The top-level items of
        the archive will appear items within this directory.
    :param only_if: A predicate of what files to include.
    :param unless: A predicate of what files to exclude.
    :param strip_components: Number of leading path components to remove when
        resolving the destination of a file from the archive. By default, the
        entire relative path from the archive root will be appended to
        ``destination``. This open will strip directory components from that
        archive path before appending it to ``destination``. If the number of
        path elements is less than or equal to ``strip_components``, the file
        will be dropped.
    :param format: The archive format. May be guessed from the file extension of
        the archive.
    :param if_exists: The action to perform if the destination directory exists.
    :param if_file_exists: The action to perform if the extraction of an
        archive member collides with an existing file in the destination.
    :param cancel: A cancellation token for the operation.
    :param on_extract: An event handler that receives an event whenever an
        archive member is extracted.
    :returns: The path referred to by ``destination``.

    .. note:: ``only_if`` is evaluated before ``unless``.
    """
    format = _get_format(Path(path), format)

    assert if_file_exists != 'merge', '"merge" is not valid for if_file_exists'

    destination = Path(destination)

    if not await _prep_expand(destination, if_exists, cancel):
        # We want to keep the contents
        return destination

    archive: Union[zipfile.ZipFile, tarfile.TarFile]
    if is_tar_format(format):
        archive = tarfile.open(path, 'r:*')
    elif format == 'zip':
        archive = zipfile.ZipFile(path, 'r')
    else:
        raise RuntimeError(f'Unknown archive format spec: {repr(format)}')

    return await _expand_archive(
        archive,
        destination,
        only_if or (lambda _: True),
        unless or (lambda _: False),
        strip_components,
        if_file_exists,
        cancel,
        on_extract,
    )



"""
Module ``dagon.core.result``
############################

Task result wrappers and status information
"""

from __future__ import annotations

import types
from dataclasses import dataclass
from typing import Any, Generic, NamedTuple, NoReturn, Tuple, Type

from .ll_dag import NodeT

ExceptionInfo = Tuple[Type[BaseException], BaseException, types.TracebackType]


class Success(NamedTuple):
    """
    Represents a successful execution of a task to completion
    """
    value: Any
    "The value that resulted from the task's execution (i.e. the return value)"


class Cancellation:
    """
    Represents that a task was cancelled before completing.
    """


class Failure(NamedTuple):
    """
    Represents that a task failed with an exception
    """
    exc_info: ExceptionInfo
    "The exception information triple that was raised while executing the task"

    @property
    def exception_type(self) -> None | Type[BaseException]:
        """The exception type associated with this failure (if applicable)"""
        return self.exc_info[0]

    @property
    def exception(self) -> None | BaseException:
        """The exception associated with this failure (if applicable)"""
        return self.exc_info[1]

    @property
    def traceback(self) -> None | types.TracebackType:
        """The exception traceback associated with this failure (if applicable)"""
        return self.exc_info[2]

    def reraise(self) -> NoReturn:
        """
        Re-raise the exception from this failure, if one is present, otherwise
        raises :class:`RuntimeError`.

        This function does not return.
        """
        exc = self.exception
        if exc is None:
            raise RuntimeError('No exception associated with this failure')
        raise exc


@dataclass(frozen=True)
class NodeResult(Generic[NodeT]):
    """
    The result of a task's execution
    """
    task: NodeT
    "The task that produced the result"
    result: Success | Cancellation | Failure
    "The actual result: A success, cancellation, or failure"



"""
Module ``dagon.task``
#####################

Mid/high-level Task APIs

"""

from __future__ import annotations

from typing import Any, Awaitable, Callable, Iterable, Type, cast, overload

from dagon.util import T, dot_kebab_name

from .dag import TaskDAG, current_dag, result_of
from .task import Dependency, Task, TaskFunction

__all__ = [
    'define',
    'define_in',
    'fn_task',
    'gather',
    'create_task_from_function',
    'Task',
    'Dependency',
    'DependsArg',
    'TaskFunction',
    'result_of',
]


class InvalidParameters(RuntimeError):
    """
    Exception raised when a task function has an invalid parameter set.
    """


DependsArg = Iterable['Task[Any] | str']
"""
Type of an dependency list for a task. Either a :class:`Task`,
a ``str``, or a sequence thereof.
"""


def iter_deps(deps: DependsArg | None, order_only_depends: DependsArg | None) -> Iterable[Dependency]:
    for dep in (deps or ()):
        d = dep if isinstance(dep, str) else dep.name
        yield Dependency(d, is_order_only=False)

    for dep in (order_only_depends or ()):
        d = dep if isinstance(dep, str) else dep.name
        yield Dependency(d, is_order_only=True)


def create_task_from_function(fn: TaskFunction[T],
                              *,
                              default: bool = True,
                              name: str | None = None,
                              depends: DependsArg = (),
                              order_only_depends: DependsArg = (),
                              doc: str | None = None,
                              cls: Type[Task[T]] = Task,
                              disabled_reason: str | None = None) -> Task[T]:
    """
    Create a :class:`Task` from the given function.

    :param fn: The function to create a task for.
    :param default: If ``True``, the task will be marked as "default"
        for execution.
    :param name: The name of the task. If not given, will use
        :func:`kebab_name` on the function's ``__name__`` to generate the name
        of the task.
    :param depends: Implicit dependencies of the task. Dependencies will
        attempt to be inferred from the signature of the given function using
        :func:`get_paramspec`, but additional dependencies can be listed using
        this parameter.
    :param order_only_depends: A sequence of dependencies that are *order-only*,
        meaning that they will not be transitively marked for execution as
        dependencies when generating the execution plan, but if they *are*
        otherwise marked, they will be required to complete successfully
        before this task is ready to be executed.
    :param doc: The help string for this task.
    :param cls: The class that should be instantiated for this task.
    :param disabled_reason: Disable the task with the given reason.

    For more parameter information, see :class:`Task`.
    """
    if name is None:
        name = dot_kebab_name(fn.__name__)

    from inspect import isclass
    if not isclass(cls) or not issubclass(cls, Task):
        raise TypeError(f'`cls` argument must be a subclass of dagon.task.Task (Got {repr(cls)})')

    doc = (fn.__doc__ if doc is None else doc) or ''

    t: Task[T] = cls(
        name=name,
        fn=fn,
        depends=list(iter_deps(
            deps=depends or (),
            order_only_depends=order_only_depends or (),
        )),
        default=default,
        doc=doc,
        disabled_reason=disabled_reason,
    )
    return t


def _task_decorator(
    dag: TaskDAG,
    *,
    default: bool,
    name: str | None,
    depends: DependsArg = (),
    order_only_depends: DependsArg = (),
    doc: str | None,
    cls: Type[Task[Any]],
    disabled_reason: str | None,
) -> Callable[[TaskFunction[T]], Task[T]]:
    def decorate_task_fn(fn: TaskFunction[T]) -> Task[T]:
        t = create_task_from_function(fn,
                                      default=default,
                                      name=name,
                                      depends=depends,
                                      doc=doc,
                                      order_only_depends=order_only_depends,
                                      cls=cls,
                                      disabled_reason=disabled_reason)
        return dag.add_task(t)

    return decorate_task_fn


def define(
    *,
    name: str | None = None,
    default: bool = False,
    depends: DependsArg = (),
    doc: str | None = None,
    order_only_depends: DependsArg = (),
    cls: Type[Task[Any]] = Task,
    disabled_reason: str | None = None,
) -> Callable[[TaskFunction[T]], Task[T]]:
    """
    Add the given task function to the default DAG.

    .. seealso:: Refer to :func:`.task_from_function`
    """
    return _task_decorator(
        current_dag(),
        default=default,
        name=name,
        depends=depends,
        order_only_depends=order_only_depends,
        doc=doc,
        cls=cls,
        disabled_reason=disabled_reason,
    )


def define_in(
    dag: TaskDAG,
    *,
    name: str | None = None,
    default: bool = False,
    depends: DependsArg = (),
    order_only_dependss: DependsArg = (),
    doc: str = '',
    cls: Type[Task[Any]] = Task,
    disabled_reason: str | None = None,
) -> Callable[[TaskFunction[T]], Task[T]]:
    """
    Create a task decorator that adds the given task to the given DAG
    """
    return _task_decorator(
        dag,
        default=default,
        name=name,
        depends=depends,
        order_only_depends=order_only_dependss,
        doc=doc,
        cls=cls,
        disabled_reason=disabled_reason,
    )


def fn_task(name: str,
            fn: Callable[[], Awaitable[T]],
            *,
            dag: TaskDAG | None = None,
            depends: DependsArg = (),
            order_only_depends: DependsArg = (),
            default: bool = False,
            doc: str | None = None,
            disabled_reason: str | None = None) -> Task[T]:
    """
    Create a task that executes a Python coroutine function.

    :param fn: The coroutine function to execute. See below.

    See :class:`.Task` for more parameter information.

    **The fn parameter**:

    The :obj:`fn` parameter must be given a coroutine function (defined with
    ``async def``) to execute. It *must* accept a :class:`.Context`
    as its first positional argument.

    The return value will be used as the result for the task.
    """
    async def _fn_trampoline() -> Any:
        return await fn()

    t = create_task_from_function(
        _fn_trampoline,
        name=name or dot_kebab_name(fn.__name__),
        default=default,
        depends=depends,
        order_only_depends=order_only_depends,
        doc=(fn.__doc__ or '') if doc is None else doc,
        disabled_reason=disabled_reason,
    )
    (dag or current_dag()).add_task(t)
    return t


@overload
def gather(name: str,
           tasks: DependsArg,
           *,
           dag: TaskDAG | None = None,
           default: bool = False,
           doc: str | None = None,
           order_only_depends: DependsArg = ()) -> Task[None]:
    ...


@overload
def gather(name: str,
           tasks: DependsArg,
           *,
           dag: TaskDAG | None = None,
           default: bool = False,
           doc: str | None = None,
           order_only_depends: DependsArg = (),
           return_value: T) -> Task[T]:
    ...


@overload
def gather(name: str,
           tasks: DependsArg,
           *,
           dag: TaskDAG | None = None,
           doc: str | None = None,
           order_only_depends: DependsArg = (),
           return_fn: Callable[[], T]) -> Task[T]:
    ...


def gather(name: str,
           tasks: DependsArg,
           *,
           dag: TaskDAG | None = None,
           default: bool = False,
           doc: str | None = None,
           order_only_depends: DependsArg = (),
           return_value: Any | None = None,
           return_fn: Callable[[], Any] | None = None) -> Task[Any]:
    """
    Declare a "gathering" task that simply depends on the given tasks but does
    nothing when it is executed.

    :param tasks: The tasks that the gather-task will depend on.
    :param return_value: Specify the result value for the gather-task, which
        will be returned from :func:`.Context.result_from` when used in
        downstream tasks.
    :param return_fn: Like :obj:`return_value`, but the function will be
        executed to generate the return value.

    .. warning::
        The result of ``return_fn()`` is returned *as-is*, with no ``await`` on
        the result. This means ``async`` functions for ``return_fn`` will
        probably not behave as expected!

    .. note::
        If neither :obj:`return_fn` nor :obj:`return_value` are specified, then
        the "result" of this task will be ``None``.

    See :func:`.define` for more parameter information.

    :return: The :class:`.Task` that was created
    """
    if return_value is not None:
        assert return_fn is None, 'Only specify one of `return_value` or `return_fn`, not both.'
        return_fn = lambda: return_value
    elif return_fn is None:
        return_fn = cast(Callable[[], T], lambda: None)
    else:
        assert callable(return_fn), f'Invalid `return_fn` object {repr(return_fn)} (must be callable)'

    # Required positional parameters
    async def _meta_fn() -> Any:
        assert return_fn
        return return_fn()

    return fn_task(
        name,
        _meta_fn,
        dag=dag,
        default=default,
        doc=doc,
        depends=tasks,
        order_only_depends=order_only_depends,
    )



"""
Classes and helpers for presenting Dag run information to the user.
"""

from __future__ import annotations

import asyncio
import contextvars
import os
import sys
import textwrap
from dataclasses import dataclass
from collections import OrderedDict
from contextlib import AsyncExitStack, ExitStack, asynccontextmanager, contextmanager
from typing import TYPE_CHECKING, AsyncIterator, Iterator, NamedTuple, Optional
from dagon.core.result import Failure, NodeResult, Success

from dagon.ext.iface import OpaqueTaskGraphView
from dagon.task.dag import OpaqueTask
from dagon.ui.events import ProgressInfo, UIEvents
from dagon.ui.message import Message, MessageType
from dagon.ui.proc import PrintProcessResultUIInfo, make_proc_info_box

from . import ansi
from .iface import I_UIExtension
from .stdio_cap import CapturedOutput, OutputAccumulator, capture_std_output
from ..util import scope_set_contextvar, typecheck

_MOVE_UP = ansi.CSI + 'A'
_MOVE_DOWN = '\n'  # ansi.CSI + 'B'
_GOTO_BOL = ansi.CSI + '1G'
_CLEAR_EOS = ansi.CSI + 'J'
_CLEAR_EOL = ansi.CSI + 'K'
_HIDE_CURSOR = ansi.CSI + '?25l'
_SHOW_CURSOR = ansi.CSI + '?25h'


@dataclass()
class _GlobalContext:
    n_total: int
    n_pending: int
    n_running: int
    n_failed: int
    n_succeeded: int
    n_cancelled: int
    stdio_cap: CapturedOutput
    stderr_acc: OutputAccumulator
    stdout_acc: OutputAccumulator


class _TaskContext(NamedTuple):
    task: OpaqueTask
    label: _FancyLabel


_GLOBAL_CTX = contextvars.ContextVar[_GlobalContext]('_GLOBAL_CTX')
_TASK_CTX = contextvars.ContextVar[_TaskContext]('_TASK_CTX')


class _FancyLabel:
    def __init__(self, name: str) -> None:
        self._name = name
        self.content: str = ''
        self.progress: Optional[float] = None

    @property
    def name(self) -> str:
        """The name string for the label"""
        return self._name


class FancyUI:
    dagon_ui_name = 'dagon.ui.fancy'
    dagon_ui_opt_name = 'fancy'

    # pylint: disable=too-many-instance-attributes
    def __init__(self) -> None:
        super().__init__()
        self._labels: dict[str, _FancyLabel] = OrderedDict()
        self._redraw_pending = False
        self._pending_output = ''
        self._pending_write = ''
        self._out_wrapper = textwrap.TextWrapper(replace_whitespace=False, drop_whitespace=False)

    @asynccontextmanager
    async def ui_global_context(self, graph: OpaqueTaskGraphView, events: UIEvents) -> AsyncIterator[None]:
        ansi.ensure_ansi_term()
        n_total = len(list(graph.all_nodes))
        async with AsyncExitStack() as st:
            cap = st.enter_context(capture_std_output())
            ctx = _GlobalContext(
                n_total=n_total,
                n_pending=n_total,
                n_running=0,
                n_failed=0,
                n_succeeded=0,
                n_cancelled=0,
                stdio_cap=cap,
                stderr_acc=OutputAccumulator(),
                stdout_acc=OutputAccumulator(),
            )
            loop = asyncio.get_event_loop()
            st.enter_context(scope_set_contextvar(_GLOBAL_CTX, ctx))
            st.callback(self._clear_on_exit)
            st.enter_context(self._hidden_cursor())
            st.enter_context(self._disable_term_echo())
            st.enter_context(events.message.connect(self._on_message))
            st.enter_context(events.progress.connect(self._on_progress))
            st.enter_context(events.task_result.connect(self._on_task_result))
            st.enter_context(events.status.connect(self._on_status))
            st.enter_context(cap.on_out.connect(lambda s: loop.call_soon_threadsafe(lambda: self._on_std_out(s))))
            st.enter_context(cap.on_err.connect(lambda s: loop.call_soon_threadsafe(lambda: self._on_std_out(s))))
            st.enter_context(events.print_process_done.connect(self._on_proc_done))
            self._redraw()
            yield

    @asynccontextmanager
    async def ui_task_context(self, task: OpaqueTask) -> AsyncIterator[None]:
        glb = _GLOBAL_CTX.get()
        glb.n_running += 1
        glb.n_pending -= 1
        label = self._labels[task.name] = _FancyLabel(task.name)
        with ExitStack() as st:
            st.enter_context(scope_set_contextvar(_TASK_CTX, _TaskContext(task, label)))
            label.content = 'Running…'
            self._schedule_redraw()
            yield
        self._labels.pop(task.name)

    def _write(self, b: str) -> None:
        self._pending_write += b

    def _commit(self) -> None:
        if os.name == 'nt':
            self._commit_win()
        else:
            self._commit_posix()

    def _commit_win(self) -> None:
        glb = _GLOBAL_CTX.get()
        glb.stdio_cap.real_stdout.write(self._pending_write)
        self._pending_write = ''

    def _commit_posix(self) -> None:
        glb = _GLOBAL_CTX.get()
        fd = glb.stdio_cap.real_stdout.fileno()
        buf = self._pending_write.encode()
        while buf:
            try:
                n_written = os.write(fd, buf)
            except BlockingIOError:
                continue
            else:
                buf = buf[n_written:]

        glb.stdio_cap.real_stdout.flush()
        self._pending_write = ''

    def _schedule_redraw(self) -> None:
        # Don't schedule multiple redraws in the same pass around the ev loop
        if not self._redraw_pending:
            asyncio.get_event_loop().call_soon(self._redraw)
        self._redraw_pending = True

    def _flush_output_lines(self) -> None:
        out = self._pending_output
        self._pending_output = ''
        self._out_wrapper.width = ansi.get_term_width() + 1
        for line in out.splitlines(keepends=True):
            if '\n' not in line:
                self._pending_output = line
                break
            wrapped_lines = self._out_wrapper.wrap(line)
            for idx, wline in enumerate(wrapped_lines):
                # Insert a blank line above and write the line into it
                self._write(f'{ansi.CSI}L{wline}{_GOTO_BOL}')
                if idx + 1 < len(wrapped_lines):
                    self._write(_MOVE_DOWN)

    def _header_text(self) -> str:
        glb = _GLOBAL_CTX.get()
        items: list[str] = []
        if glb.n_running:
            items.append(f'{glb.n_running} running')
        if glb.n_succeeded:
            items.append(ansi.bold_green(f'{glb.n_succeeded} done'))
        if glb.n_failed:
            items.append(ansi.bold_red(f'{glb.n_failed} failed'))
        if glb.n_pending:
            items.append(f'{glb.n_pending} pending')

        return ', '.join(items) + f' ({glb.n_total} total)'

    def _redraw(self) -> None:
        term_width = ansi.get_term_width()
        self._redraw_pending = False
        n_lines = len(self._labels)
        self._flush_output_lines()

        restore_height = n_lines + 2
        self._write(_MOVE_DOWN)
        max_name_width = max((len(l.name) for l in self._labels.values()), default=0)
        # Create the header
        header = self._header_text()
        header_plain = ansi.strip_escapes(header)
        header_vis_len = len(header_plain)
        vis_len_diff = len(header) - header_vis_len
        adj_term_len = term_width + vis_len_diff
        header = f'╒{f" {header} ":═^{(adj_term_len - 2)}}╕'
        bar_offset = max_name_width + 3
        header = header[:bar_offset] + '╤' + header[bar_offset + 1:]
        self._write(ansi.NORMAL + header)
        # Write the status labels
        avail_status_width = (term_width - max_name_width) - 6
        for label in self._labels.values():
            self._write(_MOVE_DOWN)
            self._write(_CLEAR_EOL)
            line = f'│ {label.name:<{max_name_width}} │ '
            progress = label.progress or 0.0
            n_progress_chars = round((avail_status_width - 1) * progress)
            st_text = label.content
            st_text = ansi.strip_escapes(st_text)
            st_text = textwrap.shorten(st_text, avail_status_width, placeholder='[…]')
            full_text = f'{st_text: <{avail_status_width}}'
            inverted = full_text[:n_progress_chars]
            normal = full_text[n_progress_chars:]
            full_text = f'{ansi.INVERSE}{inverted}{ansi.NO_INVERSE}{normal}'
            line = line + full_text + '│'
            self._write(line)
            self._write(ansi.NORMAL)
        # Write the bottom bar
        self._write(_GOTO_BOL)
        self._write(_MOVE_DOWN)
        bot = '└' + ('─' * (term_width - 2)) + '┘'
        bot = bot[:bar_offset] + '┴' + bot[bar_offset + 1:]
        self._write(bot)
        # Restore cursor
        self._write(_MOVE_UP * restore_height)
        self._write(_GOTO_BOL)

        self._commit()

    @contextmanager
    def _hidden_cursor(self) -> Iterator[None]:
        self._write(_HIDE_CURSOR)
        try:
            yield
        finally:
            self._write(_SHOW_CURSOR)

    @staticmethod
    @contextmanager
    def _disable_term_echo() -> Iterator[None]:
        if os.name == 'nt':
            yield
            return
        import termios
        term_attrs = termios.tcgetattr(sys.stdin.fileno())
        # Fourth element are the flags we want to modify
        prev_lfags = term_attrs[3]
        term_attrs[3] &= ~termios.ECHO
        termios.tcsetattr(sys.stdin.fileno(), termios.TCSANOW, term_attrs)
        try:
            yield
        finally:
            term_attrs[3] = prev_lfags
            termios.tcsetattr(sys.stdin.fileno(), termios.TCSANOW, term_attrs)

    def _clear_on_exit(self) -> None:
        # Forcibly do the final draw before reverting stdout
        self._redraw()
        self._write(_CLEAR_EOS)
        self._commit()

    def _on_task_result(self, what: NodeResult[OpaqueTask]) -> None:
        glb = _GLOBAL_CTX.get()
        glb.n_running -= 1
        if isinstance(what.result, Failure):
            glb.n_failed += 1
        elif isinstance(what.result, Success):
            glb.n_succeeded += 1
        self._write(_CLEAR_EOS)
        self._schedule_redraw()

    def _on_std_out(self, s: str) -> None:
        self._pending_output += s
        self._redraw()

    def _append_message(self, msg: Message) -> None:
        text = msg.content
        if msg.type in (MessageType.Print, MessageType.MetaPrint):
            if msg.type == MessageType.MetaPrint:
                text = ansi.bold_cyan(text)
            self._pending_output += text + '\n'
            self._schedule_redraw()
            return
        # Other logging messages should be drawn immediately with color:
        col = {
            MessageType.Error: ansi.bold_red,
            MessageType.Fatal: ansi.bold_red,
            MessageType.Warning: ansi.bold_yellow,
            MessageType.Information: ansi.bold_cyan,
        }.get(msg.type, ansi.bold_cyan)
        self._pending_output += col(text) + '\n'

    def _on_message(self, msg: Message) -> None:
        self._append_message(msg)
        # Redraw the screen immediately
        self._redraw()

    def _on_status(self, s: str) -> None:
        _TASK_CTX.get().label.content = s
        self._schedule_redraw()

    def _on_progress(self, pr: ProgressInfo) -> None:
        ctx = _TASK_CTX.get()
        progress = pr.progress
        if progress:
            progress = round(progress, 2)
        lbl = self._labels[ctx.task.name]
        if lbl.progress == progress:
            return
        lbl.progress = progress
        self._schedule_redraw()

    def _on_proc_done(self, result: PrintProcessResultUIInfo) -> None:
        for m in make_proc_info_box(result, max_width=ansi.get_term_width()):
            self._append_message(m)
        self._schedule_redraw()


if TYPE_CHECKING:
    typecheck(I_UIExtension)(FancyUI)


"""
Module for the Dagon persistent database
"""

from __future__ import annotations

import argparse
import asyncio
import dataclasses
import datetime
import enum
import hashlib
import json
import os
import sqlite3
from contextlib import ExitStack, asynccontextmanager, nullcontext
from pathlib import Path, PurePath, PurePosixPath
from typing import (Any, AsyncGenerator, AsyncIterator, ContextManager, Iterable, NamedTuple, NewType, Optional,
                    Sequence, Union, cast)

from typing_extensions import Literal, Protocol, TypeAlias

from .. import util
from ..core.result import Cancellation, Failure, NodeResult, Success
from ..event import Event, events
from ..ext.base import BaseExtension
from ..ext.iface import OpaqueTaskGraphView
from ..fs import Pathish
from ..proc import CompletedProcess
from ..task.dag import OpaqueTask
from ..util import Undefined, fixup_dataclass_docs

QueryParameter = Union[int, str, bytes, float, None]
"""Any type that is known-safe to use as a database query parameter"""


class TaskState(enum.Enum):
    """The state of a task"""
    Pending = 'pending'
    'Task is enqueued, but has not started'
    Running = 'running'
    'Task is running'
    Succeeded = 'succeeded'
    'Task executed to completion without error'
    Failed = 'failed'
    'Task raised an exception'
    Cancelled = 'cancelled'
    'Task was cancelled'


class DatabaseError(RuntimeError):
    """
    Exception thrown for database errors
    """


class MissingDatabase(DatabaseError):
    """
    Exception thrown when the requested database file is absent.
    """
    def __init__(self, path: Path) -> None:
        super().__init__(f'Missing database file at "{path}"')
        self.expected_path = path


class InvalidDatabase(DatabaseError):
    """
    Exception thrown if an existing database cannot be ready by Dagon
    """
    def __init__(self, path: Path) -> None:
        super().__init__(f'Not a Dagon database file at "{path}"')
        self.path = path


class IncorrectDatabaseVersion(DatabaseError):
    """
    Exception thrown if the database version mismatches what Dagon expects
    """
    def __init__(self, path: Path, version: str, expected_version: str) -> None:
        super().__init__(f'Database version is unsupported (Expected r{expected_version}, got r{version})')
        self.path = path
        self.version = version
        self.expected_version = expected_version


class _ProcOutputItem(Protocol):
    out: bytes
    kind: Literal['error', 'output']


def _exec_kw(db: Union[sqlite3.Connection, sqlite3.Cursor], stmt: str, **kwargs: QueryParameter) -> sqlite3.Cursor:
    return db.execute(stmt, kwargs)


def _db_connect(db_path: Pathish) -> sqlite3.Connection:
    """Open a SQLite database connection. Sets appropriate database options."""
    db = sqlite3.connect(str(db_path), isolation_level=None)
    db.row_factory = sqlite3.Row
    db.executescript(r'''
        PRAGMA foreign_keys = ON;
        PRAGMA temp_store = MEMORY;
        PRAGMA journal_mode = MEMORY;
        PRAGMA synchronous = OFF;
    ''')
    return db


_DB_SCHEMA = r'''
CREATE TABLE dagon_meta (
    hash TEXT NOT NULL
);
CREATE TABLE dagon_runs (
    run_id INTEGER PRIMARY KEY,
    time REAL NOT NULL
);
CREATE TABLE dagon_tasks (
    task_id INTEGER PRIMARY KEY,
    name TEXT NOT NULL UNIQUE
);
CREATE TABLE dagon_task_runs (
    task_run_id INTEGER PRIMARY KEY,
    task_id
        NOT NULL REFERENCES dagon_tasks
        ON DELETE CASCADE,
    run_id
        NOT NULL REFERENCES dagon_runs
        ON DELETE CASCADE,
    end_state TEXT DEFAULT 'pending',
    start_time REAL NOT NULL,
    duration REAL,
    UNIQUE(task_id, run_id) ON CONFLICT ABORT
);
CREATE TABLE dagon_task_events (
    task_event_id INTEGER PRIMARY KEY,
    task_run_id
        NOT NULL REFERENCES dagon_task_runs
        ON DELETE CASCADE,
    event NOT NULL,
    time REAL NOT NULL
);
CREATE TABLE dagon_proc_execs (
    proc_exec_id INTEGER PRIMARY KEY,
    task_run_id
        NOT NULL REFERENCES dagon_task_runs
        ON DELETE CASCADE,
    cmd NOT NULL,
    start_cwd TEXT NOT NULL,
    stdout BLOB,
    stderr BLOB,
    retc NOT NULL,
    start_time REAL NOT NULL,
    duration REAL NOT NULL
);
CREATE TABLE dagon_task_deps_rel (
    tree_id INTEGER PRIMARY KEY,
    run_id
        NOT NULL REFERENCES dagon_runs
        ON DELETE CASCADE,
    dependent TEXT NOT NULL,
    depends_on TEXT,
    UNIQUE(run_id, dependent, depends_on) ON CONFLICT REPLACE
);
CREATE TABLE dagon_persists (
    task_id REFERENCES dagon_tasks DEFAULT NULL,
    key TEXT NOT NULL,
    data TEXT NOT NULL
);
CREATE TABLE dagon_intervals (
    interval_id INTEGER PRIMARY KEY,
    task_run_id REFERENCES dagon_task_runs ON DELETE CASCADE,
    label TEXT NOT NULL,
    meta,
    start_time REAL NOT NULL,
    end_time REAL
);
'''

_DB_EXTRAS = r'''
CREATE TEMP VIEW dagon_file_data_sizes AS
    SELECT file_id,
           n,
           length(data) AS size
      FROM file_data;

CREATE TEMP VIEW dagon_file_sizes AS
    SELECT file_id,
           sum(size) AS size
      FROM dagon_file_data_sizes
  GROUP BY file_id;

CREATE TEMP VIEW dagon_sized_files AS
    SELECT * FROM files
             NATURAL JOIN dagon_file_sizes;
'''


def schema_hash() -> str:
    """
    Hash of the database schema. Used to version the database until we someday
    need to do real database migrations.
    """
    return hashlib.sha256(_DB_SCHEMA.encode('utf-8')).hexdigest()


def _create_sqlite_db(db_path: Pathish) -> sqlite3.Connection:
    db = _db_connect(db_path)
    db.executescript(_DB_SCHEMA)
    _exec_kw(db, 'INSERT INTO dagon_meta VALUES (:hash)', hash=schema_hash())
    return db


def create_new_sqlite_db(db_path: Pathish) -> sqlite3.Connection:
    """
    Create a new sqlite3 database at ``db_path``. Will remove an existing file
    if necessary.
    """
    if db_path == ':memory:':
        return _create_sqlite_db(db_path)
    db_path = Path(db_path)
    if db_path.exists():
        db_path.unlink()
    return _create_sqlite_db(db_path)


def _existing_db_schema_hash(db: sqlite3.Connection) -> str:
    try:
        for version, in _exec_kw(db, 'SELECT hash FROM dagon_meta'):
            return cast(str, version)
        return ''
    except sqlite3.OperationalError:
        return ''


def get_ready_sqlite_db(db_path: Pathish) -> sqlite3.Connection:
    """
    Get an sqlite3 connection object that is set up and ready to use to persist
    information.

    If the given DB exists and has a matching schema version, it is used as-is.
    If the DB does not exist, or the DB has the wrong schema version, a new
    database is created at ``db_path``.
    """
    db = _db_connect(db_path)
    db.execute(r'''
        CREATE TABLE IF NOT EXISTS dagon_meta (
            version INTEGER NOT NULL
        );
    ''')
    version = _existing_db_schema_hash(db)
    if version != schema_hash():
        db.close()
        db = create_new_sqlite_db(db_path)

    db.executescript(_DB_EXTRAS)
    return db


def get_existing_sqlite_db(db_path: Pathish) -> sqlite3.Connection:
    """
    Obtain an existing sqlite3 database connection. Raises an exception if the
    file does not exist or has the wrong schema version.
    """
    db_path = Path(db_path)
    if not db_path.is_file():
        raise MissingDatabase(db_path)
    db = _db_connect(db_path)
    items = list(
        _exec_kw(
            db,
            '''
            SELECT name FROM sqlite_master
            WHERE type="table" AND name='dagon_meta' LIMIT 1
            ''',
        ))
    if not items:
        raise InvalidDatabase(db_path)

    version, = util.first(_exec_kw(db, 'SELECT hash FROM dagon_meta'))
    if version != schema_hash():
        raise IncorrectDatabaseVersion(db_path, version, schema_hash())

    db.executescript(_DB_EXTRAS)
    return db


RunID = NewType('RunID', int)
TaskRowID = NewType('TaskRowID', int)
TaskRunID = NewType('TaskRunID', int)
FileID = NewType('FileID', int)
IntervalID = NewType('IntervalID', int)
ProcExecID = NewType('ProcExecID', int)

TaskRunRow: TypeAlias = 'tuple[TaskRunID, TaskRowID, RunID, int, str, float, Optional[float]]'


@dataclasses.dataclass(frozen=True)
class FileInfo:
    """
    Information about a file stored in the database
    """
    #: The Run ID that owns the file
    run_id: RunID
    #: The ID of the file object
    file_id: FileID
    #: The path to the file
    path: PurePosixPath
    #: The size of the file (in bytes). See :func:`pretty_size`
    size: int

    @property
    def pretty_size(self) -> str:
        """
        Return a prettified string representing the size of the file.
        """
        s = self.size
        for x in ('bytes', 'KiB', 'MiB', 'GiB', 'TiB'):
            if s < 1024:
                return f'{s:3.1f} {x}'
            s //= 1024
        raise RuntimeError('Value is too large. Wow.')


class Database:
    """
    Database access for persisting and obtaining data about executions.

    :param db: The SQLite database connection to use. Should have the schema
        already present.
    """
    def __init__(self, db: sqlite3.Connection) -> None:
        self._db = db
        self._transaction_lock = asyncio.Lock()

    @staticmethod
    def get_or_create(db_path: Pathish) -> 'Database':
        """
        Get or create a new database object at the given path (or ``:memory:``)
        """
        # db = create_new_sqlite_db(db_path)
        db = get_ready_sqlite_db(db_path)
        return Database(db)

    @staticmethod
    def get_existing(db_path: Pathish) -> 'Database':
        """
        Get an existing database at the named path.
        """
        db = get_existing_sqlite_db(db_path)
        return Database(db)

    @property
    def sqlite3_db(self) -> sqlite3.Connection:
        return self._db

    def new_run_id(self) -> RunID:
        """
        Create a new Run ID
        """
        c = self._db.cursor()
        _exec_kw(c, 'INSERT INTO dagon_runs (time) VALUES (:time)', time=datetime.datetime.now().timestamp())
        cur = RunID(c.lastrowid)
        return cur

    def __call__(self, q: str, **kwargs: QueryParameter) -> sqlite3.Cursor:
        """
        Execute a database query ``q`` with the given keyword argument bindings.
        """
        return _exec_kw(self._db, q, **kwargs)

    @asynccontextmanager
    async def transaction_context(self) -> AsyncGenerator[None, None]:
        async with self._transaction_lock:
            with util.recursive_transaction(self.sqlite3_db):
                yield

    def get_task_rowid(self, task: str) -> TaskRowID:
        """
        Get the ID of the named task (inserting a new row if it does not
        already exist).
        """
        for tid, in self('SELECT task_id FROM dagon_tasks WHERE name = :name', name=task):
            return TaskRowID(tid)
        c = self._db.cursor()
        _exec_kw(c, 'INSERT INTO dagon_tasks (name) VALUES (:name)', name=task)
        return TaskRowID(c.lastrowid)

    def add_task_run(self,
                     *,
                     run: RunID,
                     task: TaskRowID,
                     state: TaskState = TaskState.Pending,
                     start_time: Optional[datetime.datetime] = None) -> TaskRunID:
        """
        Create a new task run in the database.

        :param run: The run that owns the task run.
        :param task: The ID of the task (taken from :func:`get_task_rowid`)
        :param state: The initial state of the task
        :param start_time: The time at which the task started executing

        :returns: The new DB ID of the task run.
        """
        c = self._db.cursor()
        _exec_kw(
            c,
            'INSERT INTO dagon_task_runs (task_id, run_id, start_time, end_state) '
            'VALUES (:task_id, :run_id, :start_time, :state)',
            task_id=task,
            run_id=run,
            start_time=(start_time or datetime.datetime.now()).timestamp(),
            state=state.value,
        )
        trun_id = c.lastrowid
        return TaskRunID(trun_id)

    def new_interval(self, trun_id: TaskRunID, label: str, time: Optional[datetime.datetime] = None) -> IntervalID:
        """
        Create a new interval.

        :param trun_id: The task run ID, taken from :func:`add_task_run`
        :param label: The label of the interval
        :param time: The start time of the interval. Default is the current time.

        :return: The DB ID of the interval.

        .. note:: Use :func:`set_interval_end` to mark the end time of the
            interval
        """
        time = time or datetime.datetime.now()
        c = self._db.cursor()
        _exec_kw(
            c,
            r'''
            INSERT INTO dagon_intervals (task_run_id, start_time, label, meta)
            VALUES (:trun_id, :time, :label, :meta)
            ''',
            trun_id=trun_id,
            time=time.timestamp(),
            label=label,
            meta='{}',
        )
        return IntervalID(c.lastrowid)

    def set_interval_meta(self, interval: IntervalID, meta: Any) -> None:
        """
        Set a metadata value for the interval.

        :param interval: The DB ID obtained from :func:`new_interval`.
        :param meta: Any SQLite-serializable value to attach to the interval.
        """
        self(
            r'''
            UPDATE dagon_intervals
            SET meta = :meta
            WHERE interval_id = :id
            ''',
            id=interval,
            meta=json.dumps(meta),
        )

    def set_interval_end(self, interval: IntervalID, time: Optional[datetime.datetime] = None) -> None:
        """
        Mark the end time of an interval.

        :param interval: The DB ID obtained from :func:`new_interval`
        :param time: The end time of the interval. Defaults to the current time.
        """
        time = time or datetime.datetime.now()
        self(
            r'''
            UPDATE dagon_intervals
            SET end_time = :time
            WHERE interval_id = :id
            ''',
            time=time.timestamp(),
            id=interval,
        )

    def update_task_state(self,
                          trun_id: TaskRunID,
                          state: TaskState | None = None,
                          duration: datetime.timedelta | None = None) -> None:
        """
        Update the state of a task that has run.

        :param trun_id: The task run ID, obtained from :func:`add_task_run`.
        :param state: The new state. If ``None``, the state will not be changed.
        :param duration: The duration of execution. If ``None``, the duration
            will not be changed.
        """
        if state:
            self('UPDATE dagon_task_runs SET end_state = :state WHERE task_run_id = :id', state=state.value, id=trun_id)
        if duration is not None:
            self('UPDATE dagon_task_runs SET duration = :dur WHERE task_run_id = :id',
                 id=trun_id,
                 dur=duration.total_seconds())

    def store_task_event(self, trun_id: TaskRunID, event: str, time: Optional[datetime.datetime] = None) -> None:
        """
        Store an arbitrary event associated with a task.

        :param trun_id: The task run ID, obtained from :func:`add_task_run`
        :param event: String representing the event.
        :param time: The time of the event. If ``None``, uses the current time.
        """
        self(r'''
            INSERT INTO dagon_task_events (task_run_id, event, time)
            VALUES (:tid, :event, :time)
            ''',
             tid=trun_id,
             event=event,
             time=(time or datetime.datetime.now()).timestamp())

    def store_proc_execution(self, trun_id: TaskRunID, *, cmd: Sequence[str], cwd: PurePath,
                             output: Iterable[_ProcOutputItem] | None, retc: int, start_time: datetime.datetime,
                             duration: float) -> ProcExecID:
        """
        Record the execution of a subprocess.

        :param trun_id: The task run ID, obtained from :func:`add_task_run`
        :param cmd: List of strings as the command line.
        :param cwd: The path to the starting working directory of the process.
        :param stdout: The stdout output from the process.
        :param stderr: The stderr output from the process.
        :param retc: The exit code of the process.
        :param state_time: The time that the process started.
        :param duration: The duration (in seconds) of the subprocess execution.

        :return: The DB ID referencing the process execution.
        """
        q = r'''
            INSERT INTO dagon_proc_execs (
                task_run_id,
                cmd,
                start_cwd,
                stdout,
                stderr,
                retc,
                start_time,
                duration
            ) VALUES (
                :trun_id,
                :cmd,
                :cwd,
                :stdout,
                :stderr,
                :retc,
                :start_time,
                :duration
            )
        '''
        output = output or ()
        rowid = self(q,
                     trun_id=trun_id,
                     cmd=json.dumps(cmd),
                     cwd=str(cwd),
                     stdout=b''.join(o.out for o in output if o.kind == 'output'),
                     stderr=b''.join(o.out for o in output if o.kind == 'error'),
                     retc=retc,
                     start_time=start_time.timestamp(),
                     duration=duration).lastrowid
        return ProcExecID(rowid)

    def iter_files(self, *, run_id: Optional[RunID] = None) -> Iterable[FileInfo]:
        """
        Iterate over every file in the database, optionally restricted to the
        given run.

        :param run_id: If not-``None``, only yields files attached to the given
            run.
        """
        if run_id is None:
            rows = self('SELECT file_id, run_id, path, size FROM dagon_sized_files')
        else:
            rows = self('SELECT file_id, run_id, path, size FROM dagon_sized_files WHERE run_id=:r', r=run_id)

        for fid, rid, path_, size in rows:
            yield FileInfo(run_id=rid, file_id=fid, path=PurePosixPath(path_), size=size)

    def iter_file_data(self, file: Union[FileInfo, FileID]) -> Iterable[bytes]:
        """
        Iterate over the chunks of data in the file stored in the database.

        :param file: The file to read from.
        """
        if isinstance(file, FileInfo):
            file = file.file_id
        for dat, in self('SELECT data FROM file_data WHERE file_id=:fid ORDER BY n', fid=file):
            yield dat

    def add_task_dep(self, *, run_id: RunID, dependent: str, depends_on: Optional[str]) -> None:
        """
        Store information about the dependencies between tasks.

        :param run_id: The run to attach this information to.
        :param dependent: The task that has the dependency.
        :param depends_on: The task which is being depended on.
        """
        self(
            '''
            INSERT INTO dagon_task_deps_rel (run_id, dependent, depends_on)
                 VALUES (:r, :d1, :d2)
            ''',
            r=run_id,
            d1=dependent,
            d2=depends_on,
        )

    def iter_task_deps(self, run_id: RunID) -> Iterable[tuple[str, str]]:
        """
        Iterate over the dependency information for tasks in a given run.
        """
        return cast(Iterable['tuple[str, str]'],
                    self('SELECT dependent, depends_on FROM dagon_task_deps_rel WHERE run_id=:r', r=run_id))

    def load_persist(self, key: str, task_id: Optional[TaskRowID]) -> Any:
        """
        Load a persisted value from the database.

        :param key: The key of the stored value.
        :param task_id: The task that they value is attached to. If ``None``,
            it is considered a "global" persisted value.

        :return: The resulting value, or :data:`Undefined` if there is no value
            stored in the database.
        """
        data = self(
            r'''
            SELECT data
              FROM dagon_persists
              WHERE key IS :key
                    and task_id IS :task_id
            ''',
            key=key,
            task_id=task_id,
        )
        for dat_str, in data:
            return json.loads(dat_str)
        return Undefined

    def set_persist(self, key: str, task_id: Optional[TaskRowID], value: Any) -> None:
        """
        Store a persistent value in the database.

        :param key: The key under which to store the value.
        :param task_id: The task to attach the value to. If ``None``, the
            value is "global" and not attached to a task.
        :param value: The value to store. Must be JSON-serializable.
        """
        n_existing, = util.first(
            self('SELECT count(*) FROM dagon_persists WHERE key IS :key AND task_id IS :tid', key=key, tid=task_id))
        assert n_existing in (0, 1), n_existing
        if n_existing:
            self('''
                 UPDATE dagon_persists
                    SET data = :data
                  WHERE key IS :key
                        AND task_id IS :tid
                 ''',
                 data=json.dumps(value),
                 key=key,
                 tid=task_id)
        else:
            self('''
                 INSERT INTO dagon_persists (key, task_id, data)
                 VALUES (:key, :tid, :data)
                 ''',
                 key=key,
                 tid=task_id,
                 data=json.dumps(value))


@fixup_dataclass_docs
@dataclasses.dataclass()
class _AppContext():
    db_path: Path | None = None


class GlobalContext(NamedTuple):
    database: Database
    run_id: RunID


class TaskContext(NamedTuple):
    task_id: TaskRowID
    task_run_id: TaskRunID
    start_time: datetime.datetime


class _TaskContextPriv(NamedTuple):
    pub: TaskContext
    iv_stack: list[IntervalID]


class _DatabaseExt(BaseExtension[_AppContext, GlobalContext, _TaskContextPriv]):
    dagon_ext_name = 'dagon.db'
    dagon_ext_requires = ['dagon.pools', 'dagon.events']

    def add_options(self, arg_parser: argparse.ArgumentParser) -> None:
        grp = arg_parser.add_argument_group('Task Database Options')
        grp.add_argument('--db-path',
                         metavar='<path>',
                         type=Path,
                         help='Path to the database file to use. Default is ".dagon.db"',
                         default=os.environ.get('DAGON_DATABASE_PATH', '.dagon.db'))

    def handle_options(self, opts: argparse.Namespace) -> None:
        self.app_data().db_path = opts.db_path

    def app_context(self) -> ContextManager[_AppContext]:
        return nullcontext(_AppContext())

    @asynccontextmanager
    async def global_context(self, graph: OpaqueTaskGraphView) -> AsyncIterator[GlobalContext]:
        db = Database.get_or_create(self.app_data().db_path or '.dagon.db')
        rid = db.new_run_id()
        # Yield now to let the graph run
        yield GlobalContext(db, rid)
        # Link up all the dependencies
        for task in graph.all_nodes:
            for dep in graph.dependencies_of(task):
                db.add_task_dep(run_id=rid, dependent=task.name, depends_on=dep.name)

    @asynccontextmanager
    async def task_context(self, task: OpaqueTask) -> AsyncIterator[_TaskContextPriv]:
        db = self.global_data()
        now = datetime.datetime.now()
        tid = db.database.get_task_rowid(task.name)
        trun_id = db.database.add_task_run(run=db.run_id, task=tid, state=TaskState.Running, start_time=now)
        with ExitStack() as st:
            st.enter_context(events['dagon.interval-start'].connect(self._iv_start))
            st.enter_context(events['dagon.interval-end'].connect(self._iv_end))
            st.enter_context(
                events.get_or_register('dagon.proc.done', Event[CompletedProcess]).connect(self._proc_done))
            yield _TaskContextPriv(TaskContext(tid, trun_id, now), [])

    def _iv_start(self, name: str) -> None:
        iv = self.global_data().database.new_interval(self.task_data().pub.task_run_id, name)
        self.task_data().iv_stack.append(iv)

    def _iv_end(self, _: None) -> None:
        iv = self.task_data().iv_stack.pop()
        self.global_data().database.set_interval_end(iv)

    def _proc_done(self, p: CompletedProcess) -> None:
        # Store this process execution in the database
        db = self.global_data().database
        trun_id = self.task_data().pub.task_run_id
        iv = db.new_interval(trun_id, f'Subprocess {p.command}', p.start_time)
        dur = p.end_time - p.start_time
        pid = db.store_proc_execution(
            trun_id,
            cmd=p.command,
            cwd=p.cwd,
            output=p.result.output,
            retc=p.result.retcode,
            start_time=p.start_time,
            duration=dur.total_seconds(),
        )
        meta = {'process_id': pid}
        db.set_interval_meta(iv, meta)
        db.set_interval_end(iv, p.end_time)

    async def notify_result(self, result: NodeResult[OpaqueTask]) -> None:
        db = self.global_data()
        tctx = self.task_data().pub
        st = TaskState.Pending
        if isinstance(result.result, Cancellation):
            st = TaskState.Cancelled
        elif isinstance(result.result, Failure):
            st = TaskState.Failed
        else:
            assert isinstance(result.result, Success)
            st = TaskState.Succeeded
        db.database.update_task_state(tctx.task_run_id, state=st, duration=datetime.datetime.now() - tctx.start_time)


def global_context_data() -> GlobalContext | None:
    try:
        return _DatabaseExt.global_data()
    except LookupError:
        return None


def task_context_data() -> TaskContext | None:
    try:
        return _DatabaseExt.task_data().pub
    except LookupError:
        return None



"""
Module ``dagon.core.exec``
##########################

Low-level asynchronous graph execution APIs.
"""

from __future__ import annotations

import asyncio
import sys
from contextlib import AsyncExitStack, asynccontextmanager
from typing import Any, AsyncIterator, Awaitable, Callable, Generic, Mapping, cast

from ..util import Opaque, ensure_awaitable
from .ll_dag import LowLevelDAG, NodeT
from .result import Cancellation, ExceptionInfo, Failure, NodeResult, Success

NodeExecutorFunction = Callable[[NodeT], 'Awaitable[Any] | Any']
"""
A type alias of any function type used to "execute" a node. This is simply any
callable that accepts a node object and returns a value or an awaitable.
Whether the returned value is awaitable affects the behavior of the
`.SimpleExecutor` class handling that result.
"""


class SimpleExecutor(Generic[NodeT]):
    """
    A class that organizes and manages the execution of a `~.LowLevelDAG`.

    This class creates :class:`asyncio.Task` objects for tasks in the graph and
    enqueues them when ready.

    :param graph: The graph of nodes that will be executed.
    :param exec_fn: A function that is invoked to "execute" node objects.
    :param loop: The event loop that will be used during execution. If omitted,
        will use the current thread's thread loop. There *must* be a valid
        thread loop for the executor to initialize with, as it needs to prime
        the result `asyncio.Future` objects.

    .. note::
        The :obj:`graph` object will be modified by this executor's work! If you
        want to retain the original graph unmodified, use `.LowLevelDAG.copy`
        to create a copy of the graph to give to the executor.

    This class can be derived from to customize an "hook" certain events. Refer
    to the `.running_scope`, and `.do_run_node` methods, which are intended to
    be overridden for the purpose of intercepting certain graph events.

    .. note::
        It is safe to add additional nodes and edges to the graph while it is
        being executed, except for additional inbound edges to finished/running
        nodes, which is forbidden by `.LowLevelDAG.add` and guarded with an
        exception.
    """
    def __init__(self,
                 graph: LowLevelDAG[NodeT],
                 exec_fn: NodeExecutorFunction[NodeT],
                 *,
                 loop: asyncio.AbstractEventLoop | None = None) -> None:
        self.__graph = graph
        'Our graph'
        self.__exec_fn = cast(Callable[[NodeT], 'Opaque | Awaitable[Opaque]'], exec_fn)
        'The node-executing function'
        self.__running: set[asyncio.Task[NodeResult[NodeT]]] = set()
        'The set of :class:`asyncio.Task` tasks that map to currently running nodes'
        self.__futures: dict[NodeT, 'asyncio.Future[NodeResult[NodeT]]'] = {}
        'The result map for all nodes'
        self.__any_failed = False
        'Whether any nodes have failed'
        self.__started = False
        'Whether we have called on_finish'
        self.__running_scope = AsyncExitStack()
        'A context manager in scope while running the graph'
        self.__finished = False
        'Whether we have closed the running scope'

    @property
    def has_pending_work(self) -> bool:
        """
        Whether the executor has running or pending work remaining.

        .. note::
            It is possible that `.finished` is `True` while `.has_pending_work`
            is *also* `True`: If any node executions raise an exception, then
            the executor will stop enqueueing work and will declare itself to
            be finished once all running work has finished.
        """
        return self.__graph.has_ready_nodes or bool(self.__running)

    @property
    def has_running_work(self) -> bool:
        """Whether there is currently any running work"""
        return bool(self.__running)

    @property
    def any_failed(self) -> bool:
        """
        Whether any of the nodes in the graph have experienced failure/cancellation.
        """
        return self.__any_failed

    @property
    def finished(self) -> bool:
        """
        Whether the execution is finished and no more work will be queued.

        This value becomes `True` immediately after the final `NodeResult` is
        collected by `.run_some` and no more nodes are set to be executed.

        .. note::
            It is possible that `.has_pending_work` is `True` while `.finished`
            is *also* `True`: If any node executions raise an exception, then
            the executor will stop enqueueing work and will declare itself to
            be finished once all running work has finished.
        """
        return self.__finished

    def result_of(self, node: NodeT) -> Awaitable[NodeResult[NodeT]]:
        """Obtain the awaitable on the result of the given node"""
        return self.__futures[node]

    def _start_more(self) -> None:
        """
        Enqueue all ready tasks in the associated task graph.
        """
        nodes_to_start = list(self.__graph.ready_nodes)
        new_tasks = set(asyncio.get_running_loop().create_task(self.do_run_node(n)) for n in nodes_to_start)
        for n in nodes_to_start:
            self.__graph.mark_running(n)
        self.__running |= new_tasks

    async def run_some(self,
                       *,
                       interrupt: asyncio.Event | None = None,
                       timeout: float | None = None) -> Mapping[NodeT, NodeResult[NodeT]]:
        """
        Start pending nodes, and wait for one or more running nodes to finish,
        or until the `interrupt` `~asyncio.Event` is signaled, or `timeout` is
        reached.

        :param interrupt: An `~asyncio.Event` that can be used to interrupt the
            executor waiting on any nodes to finish.
        :param timeout: A timeout (in seconds) after which the call to
            `.run_some` will stop waiting on pending results and return.

        .. note:: Asserts that `.finished` is `False`

        This method will return a `set` of `.NodeResult` objects that represent
        the nodes that completed during this execution of `.run_some`. The
        returned `set` will never be empty unless `interrupt` is triggered or
        `timeout` is reached.

        .. note:: The `interrupt` event will not be cleared after we receive
            an interuption. If called in a loop with the same `~asyncio.Event`
            object, be sure to `~asyncio.Event.clear` the event before each call
            to `.run_some`.
        """
        assert not self.finished, 'run_some() called on finished executor'
        if not self.__started and self.has_pending_work:
            await self.__running_scope.enter_async_context(self.running_context())
            self.__started = True

        if not self.__any_failed:
            # There are no failures. Keep adding work
            self._start_more()

        if not self.__running:
            # There is no work to run
            return {}

        # Create a future waiting on any of the tasks
        tasks_fut = asyncio.ensure_future(asyncio.wait(self.__running, return_when=asyncio.FIRST_COMPLETED))

        # Start a future to wait on the interuption
        interrupt = interrupt or asyncio.Event()
        intr_fut = asyncio.ensure_future(interrupt.wait())

        # Wait on both the interrupt or on any of the tasks completing
        wait_both = asyncio.wait({tasks_fut, intr_fut}, return_when=asyncio.FIRST_COMPLETED, timeout=timeout)
        done = (await wait_both)[0]

        intr_fut.cancel()

        if tasks_fut not in done:
            # This occurs if we were signaled to return or if we hit our timeout
            return {}

        ret: dict[NodeT, NodeResult[NodeT]] = {}
        done, self.__running = await tasks_fut
        for t in done:
            res = await t
            self.__graph.mark_finished(res.task)
            ret[res.task] = res

        if not self.any_failed:
            # Check if any of the executions represents a failure
            self.__any_failed = any((not isinstance(r.result, Success)) for r in ret.values())

        if not self.has_pending_work or (self.any_failed and not self.has_running_work):
            self.__finished = True
            await self.__running_scope.aclose()

        return ret

    async def run_all(self, *, interrupt: asyncio.Event | None = None) -> Mapping[NodeT, NodeResult[NodeT]]:
        """
        Execute all nodes in the DAG until `.finished` is `True`, or until the
        `interrupt` `~asyncio.Event` is signaled.

        Returns a `set` of all `NodeResult` results.
        """
        ret: dict[NodeT, NodeResult[NodeT]] = {}
        while (not self.__any_failed and self.has_pending_work) or self.has_running_work:
            ret.update(await self.run_some(interrupt=interrupt))
        return ret

    def run_all_until_complete(self) -> Mapping[NodeT, NodeResult[NodeT]]:
        """
        Like `.run_all` but calls :func:`asyncio.run` with the coroutine.
        """
        return asyncio.run(self.run_all())

    def run_some_until_complete(self) -> Mapping[NodeT, NodeResult[NodeT]]:
        """
        Like `.run_some` but calls :func:`asyncio.run` with the coroutine.
        """
        return asyncio.run(self.run_some())

    @asynccontextmanager
    async def running_context(self) -> AsyncIterator[None]:
        """
        A context manager that is in scope while the graph is running, and
        exited when the graph finishes execution.

        This method is a no-op by default. It is intended to be overridden by a
        derived class.
        """
        self.__futures = {t: asyncio.get_running_loop().create_future() for t in self.__graph.all_nodes}
        yield

    def do_run_node(self, node: NodeT) -> Awaitable[NodeResult[NodeT]]:
        """
        Customization layer for handling node execution. Derived classes may
        override this method to intercept nodes before and after they are
        executed.

        .. note::
            Derived classes *should* invoke `SimpleExecutor.do_run_node()` as
            part of their derived implementation, as this method performs
            important book-keeping for tasks.

        .. note::
            This method should never raise an exception, if at all possible. If
            the execution of a node raises an exception, that failure will be
            captured at a lower layer and enclosed in a `.Failure` inside of
            the `.NodeResult` returned by `.do_run_node`.
        """
        return self._run_and_record(node)

    async def _run_and_record(self, node: NodeT) -> NodeResult[NodeT]:
        "Run the task and capture all results"
        resdat = await self._run_noraise(node)
        self.__futures[node].set_result(resdat)
        return resdat

    async def _run_noraise(self, node: NodeT) -> NodeResult[NodeT]:
        "Run the task in a way that captures all exceptions"
        try:
            retval = await ensure_awaitable(self.__exec_fn(node))
            # A successful execution with no errors. Capture the return value:
            return NodeResult(node, Success(retval))
        except asyncio.CancelledError:
            # Cancellation
            return NodeResult(node, Cancellation())
        except:  # pylint: disable=bare-except
            # An exception came from the task
            exc_info = sys.exc_info()
            assert exc_info[0]
            exc_info = cast(ExceptionInfo, exc_info)
            return NodeResult(node, Failure(exc_info))



from __future__ import annotations
from contextlib import ExitStack, contextmanager

import traceback
import warnings
from typing import Any, Callable, Dict, Generic, Iterator, Mapping, NamedTuple

from ..util import T, unused

Handler = Callable[[T], Any]
HandlerMap = Dict['ConnectionToken', Handler[T]]


class ConnectionToken(NamedTuple):
    key: int
    event: Event[Any]

    def __enter__(self) -> None:
        pass

    def __exit__(self, _unused: Any, _unused2: Any, _unused3: Any) -> None:
        unused(_unused, _unused2, _unused3)
        self.event.disconnect(self)


class Event(Generic[T]):
    """
    An object that can be used to subscribe-to or dispatch events.

    When an event is `emitted <.emit>` an associated event value will be passed.
    This same value will then be passed as the sole positional argument to all
    `handlers <.connect>`.
    """
    def __init__(self) -> None:
        self._handlers: HandlerMap[T] = {}
        self._tok_incr = 0

    def connect(self, handler: Handler[T]) -> ConnectionToken:
        """
        Register a new handler for this event. The handler must be callable with
        one positional argument, which will originate from the argument passed
        to `.emit`.

        :param handler: A function that handles events emitted by this object.

        :returns: A `.ConnectionToken` that can be used to unregister the
            event handler. See `.disconnect`.
        """
        token = ConnectionToken(self._tok_incr, self)
        self._tok_incr += 1
        self._handlers[token] = handler
        return token

    def disconnect(self, token: ConnectionToken) -> None:
        """
        Disconnect a previously connected event handler.

        :param token: A token obtained from :func:`connect`.
        """
        self._handlers.pop(token)

    def emit(self, value: T) -> None:
        """
        Emit an event, dispatching to all handlers.

        :param value: The value of the event. This object will be passed to all
            handlers.

        .. note:: Exceptions in any event handlers will be ignored with a
            warning.
        """
        # Make a copy of the handler list, in case an handler modifies it
        handlers = list(self._handlers.values())
        for h in handlers:
            try:
                h(value)
            except Exception:  # pylint: disable=broad-except
                tb = traceback.format_exc()
                warnings.warn(f'An event handler function raised an exception. It has been ignored. {tb}')


class EventMap:
    """
    A collection of named events.
    """
    def __init__(self, *, events: Mapping[str, Event[Any]] | None = None) -> None:
        self._events: dict[str, Event[Any]] = dict(events) if events is not None else {}

    def __getitem__(self, name: str) -> Event[Any]:
        """Obtain the event associated with `name`, or raise `KeyError`"""
        return self._events[name]

    def __contains__(self, name: str) -> bool:
        """Determine whether there is an event `name`"""
        return name in self._events

    def child(self) -> EventMap:
        return EventMap(events=dict(self._events))

    def register(self, name: str, event: Event[T]) -> Event[T]:
        """
        Register a new `event` in the map with `name`. If `name` is already
        present in the map, raises `NameError`.
        """
        if name in self:
            raise NameError(f'Event name "{name}" is already registered')
        self._events[name] = event
        return event

    def get(self, key: str) -> Event[Any] | None:
        """Obtain the named event, or `None` if the event is not registered"""
        return self._events.get(key)

    def get_or_register(self, name: str, factory: Callable[[], Event[T]]) -> Event[T]:
        """Obtain the named event, or register a new event created by the given factory function."""
        e = self.get(name)
        if e is None:
            e = self.register(name, factory())
        return e



from __future__ import annotations

import itertools
import shlex
import textwrap
from typing import Iterable, NamedTuple, Sequence

from typing_extensions import Literal, Protocol

from dagon.ui.message import Message, MessageType


class _ProcOutputItem(Protocol):
    out: bytes
    kind: Literal['error', 'output']


class PrintProcessResultUIInfo(NamedTuple):
    command: Sequence[str]
    retcode: int
    output: Sequence[_ProcOutputItem]


def _make_label_pair(label: str, content: str, width: int) -> Iterable[Message]:
    left = f'{label}: '
    lines = textwrap.wrap(content, width=width - 1, initial_indent=left, subsequent_indent=' ' * len(left))
    for l in lines:
        line = f'│ {l: <{width}} │'
        yield Message(line, MessageType.MetaPrint)


def _print_boxed_output(output: Iterable[_ProcOutputItem], width: int) -> Iterable[Message]:
    yield Message(f'└{f" Output: ":─<{width+2}}┘', MessageType.MetaPrint)
    for item in output:
        l = item.out.decode(errors='surrogateescape').rstrip()
        if item.kind == 'error':
            yield Message(l, MessageType.Error)
        else:
            yield Message(l, MessageType.Print)


def _box_lines(result: PrintProcessResultUIInfo, max_width: int) -> Iterable[Message]:
    cmd_plain = ' '.join(shlex.quote(s) for s in result.command)
    max_cmd = len(cmd_plain) + 10
    max_line = max((len(l.out) for l in result.output), default=0) - 5
    inner_width = max(max_cmd, max_line, 10)
    inner_width = min(inner_width, max(10, max_width - 15))
    # The top of the box:
    top = f'╒{"":═<{inner_width+2}}╕'
    yield Message(top, MessageType.MetaPrint)
    # Labels:
    pairs = (
        ('Command', cmd_plain),
        (' Exited', str(result.retcode)),
    )
    yield from itertools.chain.from_iterable(_make_label_pair(label, txt, inner_width) for label, txt in pairs)
    # The actual output:
    yield from _print_boxed_output(result.output, inner_width)
    yield Message(f'└{" End of output ":─<{inner_width+2}}┘', MessageType.MetaPrint)


def make_proc_info_box(result: PrintProcessResultUIInfo, max_width: int | None = None) -> Iterable[Message]:
    return _box_lines(result, max_width or 9999999)



from __future__ import annotations

from typing import (TYPE_CHECKING, Any, Awaitable, Callable, Generic, Iterable, NamedTuple, Sequence, Union)

from dagon.util import NoneSuch, T

IDependency = Union['Task[Any]', str]
TaskFunction = Callable[[], Awaitable[T]]


class Dependency(NamedTuple):
    """
    A dependency upon a task, which may be order-only
    """
    #: The name of the depended-on Task
    dep_name: str
    #: Whether the dependency is order-only
    is_order_only: bool


class Task(Generic[T]):
    """
    A task that wraps a callable function and contains a set of dependencies.

    :param name: The name of the task.
    :param fn: The function to execute for this task.
    :param deps: The dependencies of this task
    :param default: Whether this task should be considered "default" in task graphs of which it is a member.
    :param doc: A documentation string for this task.
    """
    def __init__(self,
                 name: str,
                 fn: TaskFunction[T],
                 depends: Iterable[Dependency],
                 default: bool = True,
                 doc: str = '',
                 disabled_reason: str | None = None):
        self._name = name
        self._doc = doc
        self._fn = fn
        self._deps = tuple(depends)
        self._is_default = default
        self._disabled_reason = disabled_reason

    @property
    def name(self) -> str:
        """The name of the task"""
        return self._name

    @property
    def doc(self) -> str:
        """The docstring associated with the task"""
        return self._doc

    @property
    def function(self) -> TaskFunction[T]:
        """The function that is associated with this task"""
        return self._fn

    @property
    def depends(self) -> Sequence[Dependency]:
        """The dependencies (by name) of this task"""
        return self._deps

    @property
    def is_disabled(self):
        """Whether this task is disabled"""
        return self.disabled_reason is not None

    @property
    def disabled_reason(self) -> str | None:
        """The explanation of why this task is disabled (or None if it is not disabled)"""
        return self._disabled_reason

    @property
    def is_default(self):
        """Whether this task is marked as a default task"""
        return self._is_default

    if not TYPE_CHECKING:

        def __call__(self, *args: Any, **kwds: Any) -> Any:
            raise RuntimeError(
                'Do not call tasks as functions! Add them as dependencies and use dagon.task.result_from()')

    def __repr__(self) -> str:
        return f'<dagon.core.Task "{self.name}">'


class InvalidTask(RuntimeError):
    """
    Exception raised when an invalid task is requested to execute.
    """
    def __init__(self, n: NoneSuch[Task[Any]]) -> None:
        super().__init__(f'Unknown task name "{n.key}"')
        self.key = n.key
        self.candidate = n.candidate


class DisabledTaskError(RuntimeError):
    """Exception thrown when an attempt is made to execute a disabled task"""
    def __init__(self, stack: Sequence[str], reason: str) -> None:
        self._task_stack = list(stack)
        root_task = stack[0]
        disabled = stack[-1]
        reqs_str = ''.join(f', which requires "{t}"' for t in stack[1:])

        super().__init__(f'Cannot mark "{root_task}"{reqs_str}, because "{disabled}" is disabled: {reason}')

    @property
    def task_stack(self) -> Sequence[str]:
        """Get the stack of tasks that led to the failure.
        The final task in the stack is the disabled task, while the first
        is the initial requested task."""
        return self._task_stack



"""
Module ``dagon.pool``
#####################

Task job pools for concurrency limits
"""
from __future__ import annotations

from asyncio import Semaphore
from contextlib import asynccontextmanager, nullcontext
from typing import (Any, AsyncContextManager, AsyncIterator, ContextManager, Hashable, NamedTuple)

from dagon.ext.iface import OpaqueTaskGraphView
from dagon.task.dag import OpaqueTask
from dagon.util import AsyncNullContext, nearest_matching

from ..ext.base import BaseExtension
from ..task.task import Task


class _PoolInfo(NamedTuple):
    name: str
    size: int


class _PoolsAppCtx(NamedTuple):
    pools: dict[str, _PoolInfo]
    assignments: dict[Hashable, str]


class _PoolsDagCtx(NamedTuple):
    pools: dict[str, Semaphore]
    assignments: dict[Hashable, Semaphore]


class _PoolsExt(BaseExtension[_PoolsAppCtx, _PoolsDagCtx, None]):
    dagon_ext_name = 'dagon.pools'

    def app_context(self) -> ContextManager[_PoolsAppCtx]:
        return nullcontext(_PoolsAppCtx({}, {}))

    def global_context(self, graph: OpaqueTaskGraphView) -> AsyncContextManager[_PoolsDagCtx]:
        appdat = self.app_data()
        semaphores = {pool.name: Semaphore(pool.size) for pool in appdat.pools.values()}
        assignments = {task: semaphores[pool_name] for task, pool_name in appdat.assignments.items()}
        return AsyncNullContext(_PoolsDagCtx(semaphores, assignments))

    @asynccontextmanager
    async def task_context(self, task: OpaqueTask) -> AsyncIterator[None]:
        gdat = self.global_data()
        sem = gdat.assignments.get(task)
        if sem is None:
            yield
            return
        async with sem:
            yield


def add(name: str, size: int) -> None:
    """
    Create a job pool in the current dag with the given name and size.

    :param name: A new unique name for the job pool. No other pool may exist
        with the same name.
    :param size: A positive non-zero integer for the number of jobs in the pool.
    """
    adat = _PoolsExt.app_data()
    if name in adat.pools:
        raise NameError(f'A pool named "{name}" is already defined')
    if size < 1:
        raise ValueError(f'Pools must have a size >=1, but "{name}" given size {size}')
    adat.pools[name] = _PoolInfo(name, size)


def assign(task: Task[Any], pool: str) -> None:
    """
    Assign the given task to the job pool with the given name. The pool must
    have been previous created using :func:`.add`. A task may only be assigned
    to one pool at a time.
    """
    adat = _PoolsExt.app_data()
    if task in adat.assignments:
        raise RuntimeError(f'Task "{task.name}" is already assigned to pool "{adat.assignments[task]}"')
    if pool not in adat.pools:
        nearest = nearest_matching(pool, adat.pools.values(), lambda p: p.name)
        if nearest:
            raise NameError(f'No pool "{pool}" is defined. (Did you mean "{nearest.name}"?)')
        raise NameError(f'No pool "{pool}" is defined')
    adat.assignments[task] = pool



from __future__ import annotations

import asyncio
from asyncio import Event
from typing import Iterable

import pytest
from dagon.core.result import NodeResult, Success

from . import exec, ll_dag
from .ll_dag import LowLevelDAG, NodeState


class CalcLengths:
    def __init__(self) -> None:
        self.strings: dict[str, int] = {}
        self.banned_strings: set[str] = set()

    def save_length(self, s: str) -> str:
        if s in self.banned_strings:
            raise RuntimeError(f'Banned string: {s}')
        self.strings[s] = len(s)
        return s


def test_calc_lengths():
    asyncio.set_event_loop(asyncio.new_event_loop())
    calc = CalcLengths()
    graph = ll_dag.LowLevelDAG(nodes=['foo', 'bar', 'baz'], edges=[('bar', 'baz'), ('foo', 'baz')])
    exe = exec.SimpleExecutor(graph.copy(), calc.save_length)
    res = exe.run_some_until_complete()
    assert exe.has_pending_work
    assert not exe.has_running_work
    assert set(res.values()) == {
        NodeResult('foo', Success('foo')),
        NodeResult('bar', Success('bar')),
    }
    assert calc.strings['foo'] == 3
    assert calc.strings['bar'] == 3
    res = exe.run_some_until_complete()
    assert set(res.values()) == {NodeResult('baz', Success('baz'))}
    assert not exe.has_pending_work
    assert not exe.has_running_work
    assert not exe.any_failed

    # Ban a string and assert that it has failed
    calc = CalcLengths()
    calc.banned_strings.add('foo')
    dup = graph.copy()
    exe = exec.SimpleExecutor(dup, calc.save_length)
    exe.run_some_until_complete()
    assert exe.has_pending_work
    assert not exe.has_running_work
    assert 'foo' not in calc.strings
    assert calc.strings['bar'] == 3
    assert exe.any_failed
    assert exe.finished


class EventWaiter:
    def __init__(self, names: Iterable[str]):
        self.events = {k: Event() for k in names}
        self.triggered_events: set[str] = set()

    async def wait_for(self, ev: str) -> str:
        await self.events[ev].wait()
        self.triggered_events.add(ev)
        return ev


@pytest.mark.asyncio
async def test_exec():
    lens = CalcLengths()
    graph = LowLevelDAG[str](nodes=['foo', 'bar', 'baz'], edges=[('foo', 'bar'), ('bar', 'baz')])
    exe = exec.SimpleExecutor(graph, lens.save_length)
    assert exe.has_pending_work
    assert graph.state_of('foo') == NodeState.Pending
    assert graph.state_of('bar') == NodeState.Pending
    assert graph.state_of('baz') == NodeState.Pending
    await exe.run_some()
    assert graph.state_of('foo') == NodeState.Finished
    assert graph.state_of('bar') == NodeState.Pending
    assert 'bar' in graph.ready_nodes
    assert exe.has_pending_work


@pytest.mark.asyncio
async def test_exec_interrupt():
    waiter = EventWaiter(['foo', 'bar', 'baz'])
    graph = LowLevelDAG[str](nodes=['foo', 'bar', 'baz'], edges=[('foo', 'bar'), ('foo', 'baz')])
    interruptor = Event()
    exe = exec.SimpleExecutor(graph, waiter.wait_for)
    assert exe.has_pending_work

    assert graph.state_of('foo') == NodeState.Pending
    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(exe.run_some(interrupt=interruptor), 0.01)

    assert graph.state_of('foo') == NodeState.Running
    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(exe.run_some(interrupt=interruptor), 0.01)

    interruptor.set()
    await asyncio.wait_for(exe.run_some(interrupt=interruptor), 1)
    assert graph.state_of('foo') == NodeState.Running

    interruptor.clear()

    # Trigger the 'foo' task to return
    waiter.events['foo'].set()
    res = await asyncio.wait_for(exe.run_some(interrupt=interruptor), 1)
    assert graph.state_of('foo') == NodeState.Finished
    assert set(res.values()) == {NodeResult('foo', Success('foo'))}
    assert exe.has_pending_work
    assert not exe.has_running_work
    assert graph.state_of('bar') == NodeState.Pending
    assert graph.state_of('baz') == NodeState.Pending

    assert 'bar' in graph.ready_nodes
    waiter.events['bar'].set()
    waiter.events['baz'].set()
    res = await asyncio.wait_for(exe.run_some(interrupt=interruptor), 1)
    assert not exe.has_pending_work
    assert not exe.has_running_work
    assert set(res.values()) == {
        NodeResult('bar', Success('bar')),
        NodeResult('baz', Success('baz')),
    }



import enum
from typing import NamedTuple


class MessageType(enum.Enum):
    """
    Types of messages that a task can emit
    """
    Information = 'info'
    Warning = 'warn'
    Error = 'error'
    Fatal = 'fatal'
    Print = 'print'
    MetaPrint = 'meta-print'


class Message(NamedTuple):
    content: str
    type: MessageType



from __future__ import annotations

import asyncio

import pytest
from dagon.core.result import NodeResult, Success

from . import dag as dag_mod
from .task import Dependency, Task


def test_simple() -> None:
    ran = 0

    async def simple_fn() -> None:
        nonlocal ran
        ran += 1
        print('Ran!')

    t1 = Task[None](name='foo', fn=simple_fn, depends=[], default=False)
    t2 = Task[None](name='bar', fn=simple_fn, depends=[Dependency('foo', is_order_only=False)], default=False)

    dag = dag_mod.TaskDAG('test')
    # Try in the wrong order
    with pytest.raises(RuntimeError, match='Unknown task name "foo"'):
        dag.add_task(t2)
    # Doesn't throw:
    dag.add_task(t1)
    # Now adding t2 will not throw
    dag.add_task(t2)

    # Run the things
    results = dag.execute(['bar'])
    results = asyncio.run(results)
    assert set(results.values()) == {
        NodeResult(t1, Success(None)),
        NodeResult(t2, Success(None)),
    }
    assert ran == 2


def test_simple_with_result():
    async def simple_two():
        return 2

    async def twice_prior() -> int:
        val: int | None = await dag_mod.result_of('foo')
        print(val)
        assert val == 2
        return 2 * val

    t1 = Task[int](name='foo', fn=simple_two, depends=[], default=False)
    t2 = Task[int](name='bar', fn=twice_prior, depends=[Dependency('foo', is_order_only=False)])
    dag = dag_mod.TaskDAG('test')
    dag.add_task(t1)
    dag.add_task(t2)
    results = asyncio.run(dag.execute(['bar']))
    assert set(results.values()) == {
        NodeResult(t1, Success(2)),
        NodeResult(t2, Success(4)),
    }



from __future__ import annotations

from dagon import event as mod


def test_simple_event():
    ev = mod.Event[int]()
    ev.emit(12)

    r: dict[str, int | None] = {'got': None}

    conn = ev.connect(lambda n: r.update({'got': n}))
    ev.emit(42)
    assert r['got'] == 42
    ev.emit(1729)
    assert r['got'] == 1729

    ev.disconnect(conn)
    ev.emit(-21)
    assert r['got'] == 1729


def test_child_map():
    emp = mod.EventMap()
    ev = emp.register('ev', mod.Event[int]())
    emp['ev'].emit(31)

    r: dict[str, int | None] = {'got': None}

    conn = ev.connect(lambda n: r.update({'got': n}))
    emp['ev'].emit(21)
    assert r['got'] == 21

    child = emp.child()
    child['ev'].emit(84)
    assert r['got'] == 84



import argparse
from contextlib import nullcontext
from typing import (TYPE_CHECKING, AsyncContextManager, Awaitable, ClassVar, ContextManager, Generic, Sequence, cast)
from dagon.core.result import NodeResult

from dagon.task.dag import OpaqueTask
from dagon.util import AsyncNullContext, ReadyAwaitable, typecheck

from .iface import AppDataT, GlobalDataT, IExtension, OpaqueTaskGraphView, TaskDataT
from .loader import ext_app_data, ext_global_data, ext_task_data


class BaseExtension(Generic[AppDataT, GlobalDataT, TaskDataT]):
    """
    Common base class for extensions, with reasonable defaults for all operations

    .. seealso:: `.IExtension`
    """
    dagon_ext_name: str
    dagon_ext_requires: ClassVar[Sequence[str]] = ()
    dagon_ext_requires_opt: ClassVar[Sequence[str]] = ()

    def app_context(self) -> ContextManager[AppDataT]:
        return nullcontext(cast(AppDataT, None))

    def global_context(self, graph: OpaqueTaskGraphView) -> AsyncContextManager[GlobalDataT]:
        return AsyncNullContext(cast(GlobalDataT, None))

    def task_context(self, task: OpaqueTask) -> AsyncContextManager[TaskDataT]:
        return AsyncNullContext(cast(TaskDataT, None))

    def add_options(self, arg_parser: argparse.ArgumentParser) -> None:
        pass

    def handle_options(self, opts: argparse.Namespace) -> None:
        pass

    @classmethod
    def app_data(cls) -> AppDataT:
        return ext_app_data(cls.dagon_ext_name)

    @classmethod
    def global_data(cls) -> GlobalDataT:
        return ext_global_data(cls.dagon_ext_name)

    @classmethod
    def task_data(cls) -> TaskDataT:
        return ext_task_data(cls.dagon_ext_name)

    def notify_result(self, result: NodeResult[OpaqueTask]) -> Awaitable[None]:
        return ReadyAwaitable(None)


if TYPE_CHECKING:
    typecheck(IExtension[int, str, bool])(BaseExtension[int, str, bool])



"""
Module ``dagon.proc``
#####################

Subprocess utilities
"""

from __future__ import annotations

import asyncio
import datetime
import json
import os
import signal
import subprocess
import warnings
from pathlib import Path, PurePath
from typing import (Any, Callable, Iterable, Mapping, NamedTuple, Sequence, Union, cast)

from typing_extensions import Literal, Protocol

from ..event import Event, events

from .. import ui
from ..event.cancel import CancellationToken, CancelLevel, raise_if_cancelled
from ..fs import Pathish
from ..task import (DependsArg, Task, TaskDAG, current_dag, iter_deps)
from ..ui import MessageType

_AioProcess = asyncio.subprocess.Process

CommandArg = Union[Pathish, str, int, float]
CommandLineElement = Union[CommandArg, Iterable['CommandLineElement']]

CommandLine = Iterable[CommandLineElement]
"""
An arbitrarily nested process command. Each element must be a `str`, `int`,
`float`, `Pathish`, or iterable of `CommandLine` itself. When spawning a process,
`.flatten_cmdline` will be used to create a flat iterable of all elements.
"""

ProcessCompletionCallback = Callable[['RunningProcess', 'ProcessResult'], None]

OutputMode = Literal['print', 'status', 'silent', 'accumulate']


class ProcessOutputItem(NamedTuple):
    out: bytes
    kind: Literal['error', 'output']


LineHandler = Callable[[ProcessOutputItem], None]


class ProcessResult(Protocol):
    """
    The type of object resulting from a finished process execution with captured
    output.
    """
    @property
    def retcode(self) -> int:
        "The exit code of the process"
        ...

    @property
    def output(self) -> Sequence[ProcessOutputItem]:
        "The output records of the subprocess"
        ...

    def stdout_json(self) -> Any:
        "Interpret the output as JSON data."
        ...


class _ProcessResultTup(NamedTuple):
    retcode: int
    output: Sequence[ProcessOutputItem]

    def stdout_json(self) -> Any:
        dat = b''.join(i.out for i in self.output if i.kind == 'output')
        return json.loads(dat)


class RunningProcess:
    """
    Represents an executing subprocess controlled by :mod:`asyncio`.

    :param proc: The subprocess handle.
    :param loop: The event loop that owns the subprocess.
    :param line_handler: A callback that is invoked for each line of output
        generated by the subprocess.
    :param on_done: A callback that is invoked when the subprocess completes.

    .. note::
        Prefer to use :func:`.spawn` to create this object rather than creating
        one by hand.
    """
    def __init__(self,
                 *,
                 proc: _AioProcess,
                 loop: asyncio.AbstractEventLoop,
                 line_handler: LineHandler | None = None,
                 on_done: ProcessCompletionCallback | None = None,
                 timeout: datetime.timedelta | None = None) -> None:
        self._loop = loop
        self._pipe = proc
        self._line_handler = line_handler
        self._on_done = on_done

        assert proc.stderr
        assert proc.stdout
        self._output: list[ProcessOutputItem] = []
        self._stderr: asyncio.Future[None] = loop.create_task(self._read(proc.stderr, 'error'))
        self._stdout: asyncio.Future[None] = loop.create_task(self._read(proc.stdout, 'output'))
        self._result: asyncio.Task[ProcessResult] = loop.create_task(self._wait_result())
        self._timer: asyncio.TimerHandle | None = None
        if timeout is not None:
            self._timer = loop.call_later(timeout.total_seconds(), self._timeout)
        self._done = False
        self._cancelled = False
        self._timed_out = False

    async def _read(self, pipe: asyncio.StreamReader, kind: Literal['error', 'output']) -> None:
        if pipe is None:
            return None

        line_acc = b''
        while 1:
            data = await pipe.read(1024)
            line_acc += data
            while True:
                nl_offset = line_acc.find(b'\n')
                if nl_offset < 0:
                    break
                line, line_acc = line_acc[:nl_offset + 1], line_acc[nl_offset + 1:]
                item = ProcessOutputItem(line, kind)
                if self._line_handler:
                    self._line_handler(item)
                self._output.append(item)
            if not data:
                break
        if line_acc:
            # Flush any remaining bytes that may not have been followed by a newline
            item = ProcessOutputItem(line_acc, kind)
            if self._line_handler:
                self._line_handler(item)
            self._output.append(item)

    async def _wait_result(self) -> ProcessResult:
        rc = await self._pipe.wait()
        await self._stdout
        await self._stderr
        self._done = True
        if self._timer:
            self._timer.cancel()
        result = _ProcessResultTup(rc, self._output)
        if self._on_done:
            self._on_done(self, result)
        if self._timed_out:
            raise TimeoutError()
        if self._cancelled:
            raise asyncio.CancelledError()
        return result

    @property
    def result(self) -> 'asyncio.Future[ProcessResult]':
        """
        The process result. Awaiting on this object will block until the
        process exits.

        :rtype: asyncio.Future[ProcessResult]
        """
        return self._result

    def send_signal(self, sig: int, to_pgrp: int = True) -> None:
        """
        Send a signal to the process or its process group.

        :param sig: The signal number to send.
        :param to_pgrp: If `True`, sends the signal to the process group. The
            process should be a group leader. This is only supported on
            Unix-like systems.
        """
        if to_pgrp and not os.name == 'nt':
            os.killpg(self._pipe.pid, sig)
        else:
            self._pipe.send_signal(sig)

    def mark_cancelled(self) -> None:
        """
        Mark the process as 'cancelled'. This will cause the process result to
        raise an `asyncio.CancelledError` regardless of the result of
        the subprocess. Note that this method does not interact with the
        subprocess in any way.

        One should prefer to use a `.CancellationToken` to cancel a running
        process.
        """
        self._cancelled = True

    def _timeout(self) -> None:
        if self._done:
            return
        try:
            self.send_signal(signal.SIGINT, to_pgrp=True)
        except ProcessLookupError:
            return
        else:
            self._timed_out = True
            self._loop.call_later(10, self._timeout_terminate)

    def _timeout_terminate(self) -> None:
        if self._done:
            return
        try:
            self._pipe.terminate()
        except ProcessLookupError:
            return
        else:
            self._loop.call_later(10, self._timeout_kill)

    def _timeout_kill(self) -> None:
        if self._done:
            return
        try:
            self._pipe.kill()
        except ProcessLookupError:
            pass

    async def communicate(self, dat: None | str | bytes = None) -> ProcessResult:
        """
        Send data to the subprocess's standard input stream and wait for the
        process to exit.

        :param dat: The input data. If given a `str`, it will be encoded with
            UTF-8 before being sent. If not given, no data will be written to
            the standard input stream. If this is not `None`, then the
            process should have been opened with `stdin` set to be a pipe. If
            the process has already exited, then no data will be written.
        """
        if isinstance(dat, str):
            dat = dat.encode('utf-8')
        if self._done:
            warnings.warn('Writing to stdin of already finished process')
        if dat is not None:
            if self._pipe.stdin is None:
                warnings.warn('Cannot write data to process: stdin was not opened as a pipe')
            else:
                self._pipe.stdin.write(dat)
                await self._pipe.stdin.drain()
                self._pipe.stdin.close()
        return await self.result


class CompletedProcess(NamedTuple):
    start_time: datetime.datetime
    end_time: datetime.datetime
    command: Sequence[str]
    cwd: Path
    result: ProcessResult
    process: RunningProcess


def arg_to_string(item: CommandArg) -> str:
    """
    Convert a command argument to a string.

    This function may look unecessary, but it only accepts certain types,
    whereas `str` can be called with anything implementing `__str__`.

    The following types are supported as command-line elements:

    - `~pathlib.PurePath` and its subclasses (including
      `~pathlib.Path`)
    - `int`
    - `float`
    - `str`
    """
    return str(item)


def flatten_cmdline(cmd: CommandLine) -> Iterable[CommandArg]:
    """
    Dagon supports nested sequences as command lines, but the subprocess APIs
    do not. This API will flatten such nested sequences into a flat iterable.

    :param cmd: The iterable of command line arguments, possibly nested.

    :raises: `TypeError` if any of the elements of the iterables is not
        a valid command-line element that Dagon supports.
        See :func:`arg_to_string`.

    .. note::
        It is not necessary to call this function when passing command lines to
        most Dagon APIs, as they will call this function themselves.
    """
    for arg in cmd:
        if isinstance(arg, (str, int, float, PurePath)):
            yield arg
        elif hasattr(arg, '__fspath__'):
            yield cast(Pathish, arg)
        elif hasattr(arg, '__iter__'):
            yield from flatten_cmdline(arg)  # type: ignore
        else:
            raise TypeError(f'Invalid command line argument element {repr(arg)}')


def plain_commandline(cmd: CommandLine) -> list[str]:
    """
    Convert a Dagon-supported nested command-line object into a `list` of `str`.
    """
    if isinstance(cmd, str):
        raise TypeError('You have tried to spawn a process using '
                        'a string as the command. Commands are not '
                        'strings! Pass a list of command line arguments '
                        'instead. NOTE: Shell-isms are NOT supported.')
    return [arg_to_string(s) for s in flatten_cmdline(cmd)]


def get_effective_env(env: Mapping[str, str] | None, *, merge: bool = True) -> dict[str, str]:
    if merge:
        final_env = os.environ.copy()
    else:
        assert env is not None, 'get_env() requires an `env` argument when merge==True'
        final_env = dict(env)
    if env and merge:
        final_env.update(env)
    return final_env


_PrintOnFinishArg = Literal['always', 'never', 'on-fail', None]


def _proc_done(cmd: Sequence[str], start_time: datetime.datetime, cwd: Path, print_output_on_finish: _PrintOnFinishArg,
               result: ProcessResult, proc: RunningProcess, next_handler: ProcessCompletionCallback) -> None:
    end_time = datetime.datetime.now()
    ev: Event[CompletedProcess] = events.get_or_register('dagon.proc.done', Event)
    ev.emit(CompletedProcess(start_time, end_time, cmd, cwd, result, proc))
    if (print_output_on_finish == 'always' or (print_output_on_finish == 'on-fail' and result.retcode != 0)):
        ui.print_process_done(ui.PrintProcessResultUIInfo(cmd, result.retcode, result.output))
    next_handler(proc, result)


async def spawn(cmd: CommandLine,
                *,
                cwd: Pathish | None = None,
                env: Mapping[str, str] | None = None,
                merge_env: bool = True,
                stdin_pipe: bool = False,
                on_output: OutputMode | LineHandler | None = None,
                on_done: ProcessCompletionCallback | None = None,
                print_output_on_finish: _PrintOnFinishArg = None,
                cancel: CancellationToken | None = None,
                timeout: datetime.timedelta | None = None) -> RunningProcess:
    """
    Spawn a subprocess and return a handle to that process.

    :param cmd: The command-line used to spawn the subprocess. Dagon supports
        nested iterables as command-line objects, which will be flattened into
        a list of strings before being used to spawn the process. Refer to
        :func:`flatten_cmdline` and :func:`arg_to_string`.
    :param cwd: The working directory for the subprocess. If not provided, will
        inherit the working directory of the Python process.
    :param env: Set the environment variables for the subprocess. **Note** that
        if `merge_env` if `False`, this will *replace* the environment
        variables, not append to the environment it would inherit from the
        parent process.
    :param merge_env: If `True`, the value of `os.environ` will be copied and
        then merged with `env` to produce the environment given to the new
        process. This has no effect if `env` is `None`.
    :param stdin_pipe: If `True`, a standard-input stream will be opened for
        the process and available on the returned `RunningProcess`
        object. If `False`, no standard-input stream will be given to the
        process.
    :param on_line: A function that will be called for each line of output
        produced by the subprocess.
    :param on_done: A callback that will be invoked when the subprocess exits.
    :param timeout: A timeout for the subprocess execution. If the timeout is
        reached, then a series of signals will be sent to the subprocess: First
        `~signal.SIGINT`, then after grace period, `~signal.SIGTERM`, then after
        another grace period `~signal.SIGKILL`. After the timeout is reached,
        awaiting on the process result will raise `TimeoutError`.
    :param cancel: A cancellation token for the child process. If a
        cancellation event is received, a `~signal.SIGINT` signal will be sent
        to the process (as if the user pressed :kbd:`ctrl+c`), and awaiting on
        the process result will raise `~asyncio.CancelledError`
    """
    if isinstance(cmd, str):
        raise TypeError('You have tried to spawn a process using '
                        'a string as the command. Commands are not '
                        'strings! Pass a list of command line arguments '
                        'instead. NOTE: Shell-isms are NOT supported.')
    # Get the cancellation token that may be in the task context
    cancel = CancellationToken.resolve(cancel)
    raise_if_cancelled(cancel)
    # Flatten the command line
    cmd_plain = plain_commandline(cmd)
    # Normalize to the actuall working dir
    cwd = Path(cwd or Path.cwd())
    # Get the subprocess environment
    env = get_effective_env(env, merge=merge_env)
    loop = asyncio.get_event_loop()
    preexec_fn = None
    might_cancel = cancel is not None or timeout is not None
    if might_cancel and os.name != 'nt':
        # We might send a signal to this process via `cancel`, and we
        # don't want the controlling shell to send it `SIGINT` if the user
        # presses ^C.
        #
        # Make the process a session leader of a new session, which will have
        # no terminal and therefore no one to send it ^C.
        preexec_fn = os.setpgrp

    # Normalize on_done to a do-nothing handler
    on_done = on_done or (lambda p0, p1: None)
    prev_on_done = on_done
    start_time = datetime.datetime.now()
    cwd_ = cwd
    if print_output_on_finish is None:
        if on_output == 'accumulate':
            print_output_on_finish = 'always'
        else:
            print_output_on_finish = 'on-fail'
    # Call _proc_done when the process completes
    on_done = lambda p0, p1: _proc_done(cmd_plain, start_time, cwd_, print_output_on_finish, p1, p0, prev_on_done)

    on_line = as_line_handler(on_output)

    # Spawn that subprocess!
    proc = await asyncio.create_subprocess_exec(
        *cmd_plain,
        cwd=cwd,
        env=env,
        stdin=subprocess.PIPE if stdin_pipe else subprocess.DEVNULL,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        preexec_fn=preexec_fn,
    )

    ret = RunningProcess(proc=proc, loop=loop, line_handler=on_line, on_done=on_done, timeout=timeout)
    if cancel is not None:

        def cancel_it(_lvl: CancelLevel) -> None:
            if os.name != 'nt':
                try:
                    ret.send_signal(signal.SIGINT, to_pgrp=True)
                except ProcessLookupError:
                    # The process is already dead. We're okay.
                    pass
            ret.mark_cancelled()

        token = cancel.connect(cancel_it)
        ce = cancel
        ret.result.add_done_callback(lambda _: ce.disconnect(token))
    return ret


class _UpdateStatus:
    def __call__(self, line: ProcessOutputItem) -> None:
        ui.status(line.out.decode(errors='?'))


class _LogLine:
    def __call__(self, line: ProcessOutputItem) -> None:
        ui.print(line.out.decode(errors='?'), type=MessageType.Error if line.kind == 'error' else MessageType.Print)


PRINT_OUTPUT: LineHandler = _LogLine()
UPDATE_STATUS_OUTPUT: LineHandler = _UpdateStatus()


async def run(cmd: CommandLine,
              *,
              cwd: Pathish | None = None,
              env: Mapping[str, str] | None = None,
              merge_env: bool = True,
              stdin: None | str | bytes = None,
              check: bool = True,
              on_output: LineHandler | OutputMode | None = None,
              on_done: ProcessCompletionCallback | None = None,
              print_output_on_finish: _PrintOnFinishArg = None,
              cancel: CancellationToken | None = None,
              timeout: datetime.timedelta | None = None) -> ProcessResult:
    """
    Execute a subprocess and wait for the result.

    :param check: If `True`, the exit code of the process will be checked. If
        it is non-zero, then this function will raise a
        `CalledProcessError` exception with the output information for
        the subprocess.
    :param stdin: A `bytes` or `str` object. A `str` object will be
        encoded as UTF-8 `bytes` object. The `bytes` will be sent to the
        standard-input stream of the subprocess. If `stdin` is not provided,
        the subprocess will not have a standard-input stream.

    See :func:`.spawn` for more parameter information.
    """
    proc = await spawn(
        cmd,
        cwd=cwd,
        env=env,
        merge_env=merge_env,
        stdin_pipe=stdin is not None,
        on_output=on_output,
        on_done=on_done,
        print_output_on_finish=print_output_on_finish,
        cancel=cancel,
        timeout=timeout,
    )
    result = await proc.communicate(stdin)
    if check:
        if result.retcode != 0:
            stdout = b''.join(o.out for o in result.output if o.kind == 'output')
            stderr = b''.join(o.out for o in result.output if o.kind == 'error')
            raise subprocess.CalledProcessError(
                result.retcode,
                plain_commandline(cmd),
                output=stdout,
                stderr=stderr,
            )
    return result


def cmd_task(name: str,
             cmd: CommandLine,
             *,
             dag: TaskDAG | None = None,
             cwd: Pathish | None = None,
             doc: str = '',
             on_output: LineHandler | OutputMode | None = 'accumulate',
             print_output_on_finish: _PrintOnFinishArg = None,
             check: bool = True,
             default: bool = False,
             depends: DependsArg = (),
             env: dict[str, str] | None = None,
             order_only_depends: DependsArg = (),
             disabled_reason: str | None = None) -> Task[ProcessResult]:
    """
    Declare a task that executes the given command. See
    :func:`dagon.task.define` and :func:`run` for parameter information.

    :return: The `.Task` that was created
    """
    cmd = list(cmd)
    assert cmd, '`cmd` must not be empty for declare_cmd_task'

    async def _command_runner() -> ProcessResult:
        return await run(
            cmd,
            cwd=cwd,
            check=check,
            env=env,
            on_output=on_output,
            print_output_on_finish=print_output_on_finish,
        )

    dag = dag or current_dag()

    if isinstance(depends, (Task, str)):
        depends = [depends]

    depends_ = iter_deps(depends, order_only_depends)

    t = Task[ProcessResult](name=name,
                            fn=_command_runner,
                            depends=depends_,
                            default=default,
                            doc=doc,
                            disabled_reason=disabled_reason)
    dag.add_task(t)
    return t


def as_line_handler(l: LineHandler | OutputMode | None) -> LineHandler | None:
    if l is None:
        return None
    if callable(l):
        return l
    if l in ('accumulate', 'silent'):
        return None
    if l == 'print':
        return PRINT_OUTPUT
    if l == 'status':
        return UPDATE_STATUS_OUTPUT
    assert False, f'Invalid value for line_handler/on_output: {l!r}'



"""
Inspect a Dagon execution as an interactive timeline
"""

from __future__ import annotations

import base64
import json
from pathlib import Path
from pkg_resources import resource_filename
import itertools
from typing import Any, Mapping, Sequence, cast

from dagon.db import Database, RunID, TaskRunID
from dagon.util import first


def read_resource(filepath: str) -> bytes:
    return Path(resource_filename('dagon', filepath)).read_bytes()


def _gen_targets(db: Database, rid: RunID) -> Sequence[Mapping[str, Any]]:
    target_rows = db(
        r'''
        SELECT *
        FROM dagon_task_runs
        JOIN dagon_tasks USING(task_id)
        WHERE run_id=:rid
        ''',
        rid=rid,
    )
    return [dict(zip(row.keys(), row)) for row in target_rows]


def _gen_intervals(db: Database, rid: RunID) -> Mapping[TaskRunID, Sequence[Mapping[Any, Any]]]:
    rows = db(
        r'''
        SELECT task_run_id,
               iv.start_time,
               end_time - iv.start_time AS duration,
               meta,
               label
          FROM dagon_intervals AS iv
          JOIN dagon_task_runs USING (task_run_id)
      ORDER BY task_run_id, iv.start_time
        ''',
        rid=rid,
    )
    items = {
        trun_id: [dict(zip(row.keys(), row)) for row in intervals]
        for trun_id, intervals in itertools.groupby(rows, key=lambda row: cast(TaskRunID, row[0]))
    }
    return items


def _gen_procs(db: Database, rid: RunID) -> Mapping[str, Any]:
    rows = db(
        r'''
        SELECT proc_exec_id,
               cmd,
               start_cwd,
               stdout,
               stderr,
               retc,
               dagon_proc_execs.start_time AS start_time,
               dagon_proc_execs.duration AS duration
            FROM dagon_proc_execs
            JOIN dagon_task_runs USING (task_run_id)
            WHERE run_id = :rid
        ''',
        rid=rid,
    )
    ret = {row[0]: dict(zip(row.keys(), row)) for row in rows}
    for _, row in ret.items():
        row.update({
            'stdout': base64.b64encode(row['stdout'] or b'').decode('ascii'),
            'stderr': base64.b64encode(row['stderr'] or b'').decode('ascii'),
        })
    return ret


def _generate_json_data(db: Database, rid: RunID) -> bytes:
    targets = _gen_targets(db, rid)
    intervals = _gen_intervals(db, rid)
    procs = _gen_procs(db, rid)
    start_time, = first(db(r'''SELECT time FROM dagon_runs WHERE run_id=:rid''', rid=rid))
    data = {
        'targets': targets,
        'start_time': start_time,
        'intervals': intervals,
        'proc_execs': procs,
    }
    buf = json.dumps(data).encode('utf-8')
    buf = buf.replace(b'\\', b'\\\\')
    buf = buf.replace(b'`', b'\\`')
    return buf


def generate(db: Database, rid: RunID) -> bytes:
    html = read_resource('inspect/tl.html')
    js = read_resource('inspect/tl.js')
    css = read_resource('inspect/tl.css')

    html = html.replace(b'[[TL_SCRIPT]]', js)
    html = html.replace(b'{{TL_CSS}}', css)
    html = html.replace(b'[[TL_DATA]]', _generate_json_data(db, rid))

    return html



from __future__ import annotations

import argparse
import contextvars
import sys
import warnings
from contextlib import (AsyncExitStack, ExitStack, asynccontextmanager, contextmanager)
from typing import (Any, AsyncContextManager, AsyncIterator, Callable, Iterable, Iterator, NamedTuple, Sequence, cast,
                    overload)

from dagon.core import ll_dag
from dagon.core.result import NodeResult
from dagon.task.dag import OpaqueTask
from dagon.util import Opaque, scope_set_contextvar

from .iface import AppDataT, GlobalDataT, IExtension, TaskDataT

if sys.version_info < (3, 8):
    from importlib_metadata import EntryPoint, entry_points
else:
    from importlib.metadata import EntryPoint, entry_points


class _ExtDataContext(NamedTuple):
    by_name: dict[str, Opaque]


class ExtensionLoadWarning(RuntimeWarning):
    pass


# The per-context per-extension extension data
_CTX_APP_DATA = contextvars.ContextVar[_ExtDataContext]('_CTX_APP_DATA')
_CTX_GLOBAL_DATA = contextvars.ContextVar[_ExtDataContext]('_CTX_GLOBAL_DATA')
_CTX_TASK_DATA = contextvars.ContextVar[_ExtDataContext]('_CTX_TASK_DATA')


def _get_ext_data(cvar: contextvars.ContextVar[_ExtDataContext], ext: str | IExtension[Any, Any, Any]) -> Any:
    if not isinstance(ext, str):
        ext = ext.dagon_ext_name
    ctx = cvar.get()
    try:
        return ctx.by_name[ext]
    except KeyError as e:
        raise RuntimeError(f'No extension "{ext}" is loaded in the requested context') from e


@overload
def ext_app_data(ext: IExtension[AppDataT, Any, Any]) -> AppDataT:
    ...


@overload
def ext_app_data(ext: str) -> Any:
    ...


def ext_app_data(ext: str | IExtension[Any, Any, Any]) -> Any:
    """
    Obtain the application-level context data for the given extension.

    .. note:: The extension must be loaded and we must be within an application context.
    """
    return _get_ext_data(_CTX_APP_DATA, ext)


@overload
def ext_global_data(ext: IExtension[Any, GlobalDataT, Any]) -> GlobalDataT:
    ...


@overload
def ext_global_data(ext: str) -> Any:
    ...


def ext_global_data(ext: str | IExtension[Any, Any, Any]) -> Any:
    """
    Obtain the DAG-global context data for the given extension.

    .. note:: The extension must be loaded and we must be within a DAG exection context.
    """
    return _get_ext_data(_CTX_GLOBAL_DATA, ext)


@overload
def ext_task_data(ext: IExtension[Any, Any, TaskDataT]) -> TaskDataT:
    ...


@overload
def ext_task_data(ext: str) -> Any:
    ...


def ext_task_data(ext: str | IExtension[Any, Any, Any]) -> Any:
    """
    Obtain the task-level context data for the given extension.

    .. note:: The extension must be loaded and we must be within a task execution context.
    """
    return _get_ext_data(_CTX_TASK_DATA, ext)


_OpaqueExt = IExtension[Opaque, Opaque, Opaque]
"An opaque-typed extension"


def get_entry_points(group: str) -> Iterable[EntryPoint]:
    if sys.version_info < (3, 8):
        return cast(Iterable[EntryPoint], entry_points(group=group))  # type: ignore
    elif sys.version_info < (3, 10):
        return cast(Iterable[EntryPoint], entry_points().get(group))  # type: ignore
    else:
        return cast(Iterable[EntryPoint], entry_points().select(group=group))  # type: ignore


class ExtLoader:
    """
    An extension loader and manager.

    All extensions should be loaded into this class, and this class can be used
    to control them as a group in the appropriate order.
    """
    def __init__(self) -> None:
        self._loaded: dict[str, _OpaqueExt] = {}
        "Active extensions by name"

    def load(self, ext: IExtension[Any, Any, Any]) -> None:
        """
        Load the given extension into the loader context. There must not already
        be an extension loaded by the same name.

        If the given extension has and extension dependencies, those will need
        to be loaded too before the loader can be used, but do not necessarily
        need to be loaded in order.
        """
        if ext.dagon_ext_name in self._loaded:
            raise NameError(f'Extension "{ext.dagon_ext_name}" is already loaded')
        self._loaded[ext.dagon_ext_name] = ext

    def _iter_order(self) -> Sequence[_OpaqueExt]:
        """Return a sequence of all extensions in appropriate dependency order"""
        return tuple(self._iter_order_1())

    def _iter_order_1(self) -> Iterable[_OpaqueExt]:
        visited: set[str] = set()
        for e in self._loaded.values():
            yield from self._iter_order_2(e, visited)

    def _iter_order_2(self, ext: _OpaqueExt, visited: set[str]) -> Iterable[_OpaqueExt]:
        # First load all of the requirements:
        for r in ext.dagon_ext_requires:
            # Skip if we have already loaded it
            if r in visited:
                continue
            if r not in self._loaded:
                # Raise an error if it has not been registered
                raise RuntimeError(f'No extension "{r}" is registered, but is required by "{ext.dagon_ext_name}"')
            yield from self._iter_order_2(self._loaded[r], visited)
        # And optional requirements:
        for r in ext.dagon_ext_requires_opt:
            if r in visited:
                continue
            if r in self._loaded:
                yield from self._iter_order_2(self._loaded[r], visited)
        # Now yield the actual one we are visiting:
        if ext.dagon_ext_name not in visited:
            yield ext
            visited.add(ext.dagon_ext_name)

    @asynccontextmanager
    async def _enter_context(
        self,
        cvar: contextvars.ContextVar[_ExtDataContext],
        proj: Callable[[_OpaqueExt], AsyncContextManager[Opaque]],
    ) -> AsyncIterator[None]:
        """
        Enter the async context, and store the context value in the given contextvar.

        Each extension will be entered using the ``proj`` projection function,
        which should map an extension to an associated async context manager.
        """
        async with AsyncExitStack() as st:
            data = _ExtDataContext({})
            st.enter_context(scope_set_contextvar(cvar, data))
            for p in self._iter_order():
                dat = await st.enter_async_context(proj(p))
                data.by_name[p.dagon_ext_name] = dat
            yield

    @contextmanager
    def app_context(self) -> Iterator[None]:
        """
        Enter the application-level context for all extensions. This is a non-async
        context.
        """
        with ExitStack() as st:
            data = _ExtDataContext({})
            st.enter_context(scope_set_contextvar(_CTX_APP_DATA, data))
            for p in self._iter_order():
                dat = st.enter_context(p.app_context())
                data.by_name[p.dagon_ext_name] = dat
            yield

    def global_context(self, graph: ll_dag.DAGView[OpaqueTask]) -> AsyncContextManager[None]:
        """Enter the DAG-global context for the given task graph"""
        return self._enter_context(_CTX_GLOBAL_DATA, lambda p: p.global_context(graph))

    def task_context(self, task: OpaqueTask) -> AsyncContextManager[None]:
        """Enter a task-local context for the given task"""
        return self._enter_context(_CTX_TASK_DATA, lambda p: p.task_context(task))

    async def notify_result(self, result: NodeResult[OpaqueTask]) -> None:
        """Notify all extensions of the result of a task"""
        for p in self._iter_order():
            await p.notify_result(result)

    def add_options(self, parser: argparse.ArgumentParser) -> None:
        """
        Allow all extensions to augment the command-line argument parser.
        """
        for p in self._loaded.values():
            p.add_options(parser)

    def handle_options(self, args: argparse.Namespace) -> None:
        """
        Allow all extensions to handle the result of command-line argument
        parsing.
        """
        for p in self._loaded.values():
            p.handle_options(args)

    @classmethod
    def default(cls) -> ExtLoader:
        """
        Obtain the default extension loaded loaded with the extensions
        defined by the "dagon.extensions" entrypoint. This will instantiate and
        load all extensions. Extensions that fail to load will be discarded with
        a warning message.
        """
        ret = ExtLoader()
        eps = get_entry_points('dagon.extensions')
        for ep in eps:
            cls._try_load_one(ret, ep)
        # Call the order function. This will raise an exception if any extensions
        # are loaded without their requirements having been loaded first.
        ret._iter_order()
        return ret

    @staticmethod
    def _try_load_one(ldr: ExtLoader, ep: EntryPoint) -> None:
        "Attempt to load one extension from the given EntryPoint"
        ep_name: str = getattr(ep, 'name', '<unnamed>')
        try:
            cls: _OpaqueExt | object = ep.load()
        except BaseException as e:
            warnings.warn(f"An exception occurred while trying to load extension {ep!r}: {e}",
                          ExtensionLoadWarning,
                          source=e)
            return
        # Be very careful now...
        # ext_name is required:
        if not hasattr(cls, 'dagon_ext_name'):
            warnings.warn(f'Extension object {cls!r} does not have a "dagon_ext_name" attribute', ExtensionLoadWarning)
            return
        ext_name = getattr(cls, 'dagon_ext_name')
        if not isinstance(ext_name, str):
            warnings.warn(f'Extension object {cls!r} has a non-string "dagon_ext_name": {ext_name!r}',
                          ExtensionLoadWarning)
            return
        # dagon_ext_name must match the entrypoint name:
        if ext_name != ep_name:
            warnings.warn(
                f'Extension object {cls!r} has a mismatching extension name. Expected "{ep_name}", but got "{ext_name}"',
                ExtensionLoadWarning)
            return
        # The extension must be a callable object:
        if not callable(cls):
            warnings.warn(f'Extension entrypoint {cls!r} is not callable', ExtensionLoadWarning)
            return
        # Attempt to create an instance now:
        try:
            inst: Any = cls()
        except BaseException as e:
            warnings.warn(f'Instantiating/calling extension loader [{cls!r}] resulted in an exception: {e}',
                          ExtensionLoadWarning,
                          source=e)
            return
        # Check that the extension implements the interface using a type-checkable Protocol:
        if not isinstance(inst, IExtension):
            warnings.warn(f'Generated extension object {inst!r} does not implement the necessary interface')
            return
        # Got it:
        ldr.load(cast(_OpaqueExt, inst))



"""
Module ``dagon.tool.main``
##########################
"""

from __future__ import annotations

import argparse
import asyncio
import importlib
import importlib.util
import logging
import os
import shlex
import subprocess
import sys
import textwrap
import traceback
import warnings
from pathlib import Path
from typing import Iterable, NoReturn, Optional, Sequence, cast

import pkg_resources

from ..core.result import Cancellation, Failure
from ..ext.exec import ExtAwareExecutor
from ..ext.loader import ExtLoader
from ..task.dag import TaskDAG, populate_dag_context
from ..task.task import DisabledTaskError, InvalidTask
from ..tool.args import ParsedArgs, get_argparser

_g_loading_dag = False


def _format_docstr(s: str) -> Iterable[str]:
    paragraphs = textwrap.dedent(s).split('\n\n')
    for par in paragraphs:
        for line in textwrap.wrap(par):
            yield line.strip()
        yield ''  # Empty line after paragraph


def _list_tasks(dag: TaskDAG, with_docs: bool = True) -> int:
    for t in dag.tasks:
        name_line = t.name
        if t.depends:
            name_line += ' : ' + ' '.join(shlex.quote(dep.dep_name) for dep in t.depends)
        print(name_line)
        if with_docs and t.doc:
            for line in _format_docstr(t.doc):
                print(f'    {line}')
        if t.is_disabled:
            print(f'    [DISABLED: {t.disabled_reason}]')
    return 0


def _configure_logging(level: int) -> None:
    logging.basicConfig(level=level, format='[dagon] %(message)s')


def _print_failure(task_name: str, fail: Failure) -> None:
    if isinstance(fail.exception, asyncio.CancelledError):
        # Don't print exceptions which are just task cancellations
        return

    head = f'Task "{task_name}" FAILED'

    term_width = int(os.environ.get('COLUMNS', '80'))

    if isinstance(fail.exception, subprocess.CalledProcessError):
        print(f'{head}: Subprocess execution returned non-zero', file=sys.stderr)
        traceback.print_exception(fail.exception_type, fail.exception, fail.traceback)
        print(f'{" Process output ":=^{term_width}}', file=sys.stderr)
        print((fail.exception.stdout or b'').decode(), file=sys.stderr, end='')
        print((fail.exception.stderr or b'').decode(), file=sys.stderr, end='')
        print(f'{" End Output ":^^{term_width}}', file=sys.stderr)
    else:
        print(f'{head}: Encountered an exception during execution:', file=sys.stderr)
        traceback.print_exception(fail.exception_type, fail.exception, fail.traceback)


def run_with_args(dag: TaskDAG, exts: ExtLoader, args: ParsedArgs, default_tasks: Optional[Sequence[str]]) -> int:
    if args.list_tasks:
        return _list_tasks(dag, with_docs=not args.no_doc)

    _configure_logging(logging.DEBUG if args.debug else logging.INFO)

    tasks = list(args.tasks or default_tasks or (t.name for t in dag.tasks if t.is_default))
    if not tasks:
        warnings.warn('No tasks are set to be executed')

    try:
        ll_dag = dag.low_level_graph(tasks)
    except DisabledTaskError as e:
        print(f'Unable to create the execution plan: {e}', file=sys.stderr)
        return 1
    except InvalidTask as e:
        if e.candidate:
            print(f'No such task "{e.key}". Did you mean "{e.candidate.name}" ?', file=sys.stderr)
        else:
            print(f'No such task "{e.key}".', file=sys.stderr)
        return 1

    exe = ExtAwareExecutor(exts, ll_dag, catch_signals=True, fail_cancels=args.fail_cancels)

    results = exe.run_all_until_complete()
    failed = False
    for item in results.values():
        if isinstance(item.result, Failure):
            _print_failure(item.task.name, item.result)
            failed = True
        elif isinstance(item.result, Cancellation):
            print(f'Task "{item.task.name}" was cancelled')
            failed = True
        else:
            # Nothing to print for successes
            pass

    return 1 if failed else 0


def run_for_dag(dag: TaskDAG,
                exts: ExtLoader,
                *,
                argv: Optional[Sequence[str]] = None,
                default_tasks: Optional[Sequence[str]] = None) -> int:
    argv = argv if argv is not None else sys.argv[1:]
    parser = get_argparser(exts)
    args = cast(ParsedArgs, parser.parse_args(argv))
    exts.handle_options(cast(argparse.Namespace, args))
    return run_with_args(dag, exts, args, default_tasks)


def get_extensions() -> ExtLoader:
    return ExtLoader.default()


class MainArgParseResults(ParsedArgs):
    file: Optional[str]
    module: Optional[str]
    dir: Optional[Path]
    version: bool


def main(argv: Sequence[str]) -> int:
    exts = get_extensions()
    with exts.app_context():
        return _main(argv, exts)


def _main(argv: Sequence[str], exts: ExtLoader) -> int:
    parser = get_argparser(exts)
    parser.add_argument('-f', '--file', help='The file that defines the tasks')
    parser.add_argument('-m', '--module', help='The module to import to define the tasks')
    parser.add_argument('--dir',
                        help='Add the given directory to the import/search path for finding tasks',
                        default=Path.cwd(),
                        type=Path)
    parser.add_argument(
        '--version',
        help='Display the version and exit',
        action='store_true',
    )
    args = cast(MainArgParseResults, parser.parse_args(argv))

    if args.version:
        version = pkg_resources.get_distribution('dagon').version
        print(version)
        return 0

    def_file = Path(args.file).absolute() if args.file else None
    def_mod: Optional[str] = args.module

    if def_mod and def_file:
        print('`--module` and `--file` cannot be used together', file=sys.stderr)
        return 2

    if not def_mod and not def_file:
        def_mod = 'dag'

    global _g_loading_dag  # pylint: disable=global-statement
    _g_loading_dag = True
    prev_path = list(sys.path)
    try:
        sys.path.insert(0, str(args.dir))
        dag = TaskDAG('<main>')
        if def_mod:
            with populate_dag_context(dag):
                try:
                    if def_mod in sys.modules:
                        importlib.reload(sys.modules[def_mod])
                    else:
                        importlib.import_module(def_mod)
                except ModuleNotFoundError:
                    traceback.print_exc(file=sys.stderr)
                    print(f'Failed to import the "{def_mod}" module for task definitions', file=sys.stderr)
                    return 1
        else:
            spec = importlib.util.spec_from_file_location('<dag-definition>', str(def_file))
            if not spec:
                print('Unable to find task definitions', file=sys.stderr)
                return 1
            mod = importlib.util.module_from_spec(spec)
            with populate_dag_context(dag):
                spec.loader.exec_module(mod)  # type: ignore
    finally:
        _g_loading_dag = False
        sys.path = prev_path

    exts.handle_options(cast(argparse.Namespace, args))
    return run_with_args(dag, exts, args, default_tasks=[])


# def execute_default(*, dag: TaskDAG | None = None, default: Sequence[str] | None = None) -> None:
#     if _g_loading_dag:
#         warnings.warn(
#             'Call to execute_default() while running main Dagon executable. execute_default() will do nothing.')
#         return
#     dag = dag or current_dag()
#     rc = run_for_dag(dag, argv=sys.argv[1:], default_tasks=default or [])
#     sys.exit(rc)


def start() -> NoReturn:
    sys.exit(main(sys.argv[1:]))



from __future__ import annotations

from contextlib import AsyncExitStack, asynccontextmanager
from typing import AsyncIterator

from ..core import ll_dag
from ..core.result import NodeResult
from ..task.dag import OpaqueTask, TaskExecutor
from ..util import Opaque
from .loader import ExtLoader


class ExtAwareExecutor(TaskExecutor[Opaque]):
    """
    An extension-aware task executor. Executes a task graph and loads the appropriate
    extension contexts during execution
    """
    def __init__(self, exts: ExtLoader, graph: ll_dag.LowLevelDAG[OpaqueTask], *, catch_signals: bool,
                 fail_cancels: bool) -> None:
        super().__init__(graph, self.__exec_task, catch_signals=catch_signals, fail_cancels=fail_cancels)
        self._exts = exts

    async def __exec_task(self, task: OpaqueTask) -> Opaque:
        return await task.function()

    @asynccontextmanager
    async def running_context(self) -> AsyncIterator[None]:
        async with AsyncExitStack() as st:
            await st.enter_async_context(self._exts.global_context(ll_dag.DAGView(self.graph)))
            await st.enter_async_context(super().running_context())
            yield

    async def do_run_task(self, node: OpaqueTask) -> NodeResult[OpaqueTask]:
        async with self._exts.task_context(node):
            result = await super().do_run_task(node)
            await self._exts.notify_result(result)
            return result



from contextlib import asynccontextmanager
from pathlib import Path
from typing import AsyncContextManager, AsyncGenerator, AsyncIterator
import pytest

from dagon import storage as mod
from ..db import Database
from ..fs import Pathish
from dagon import fs


async def _read_file(store: mod.IFileStorage, name: Pathish) -> bytes:
    return await mod.recover(name, from_=store)


class CaseSet_IFileStorage:
    @pytest.fixture
    @pytest.mark.asyncio
    async def store(self, tmp_path: Path) -> AsyncGenerator[mod.IFileStorage, None]:
        async with self.do_create_storage(tmp_path) as f:
            yield f

    def do_create_storage(self, tmp_path: Path) -> AsyncContextManager[mod.IFileStorage]:
        raise NotImplementedError

    @pytest.mark.asyncio
    async def test_simple_open(self, store: mod.IFileStorage):
        async with store.open_file_writer('test.txt', if_exists='replace') as wr:
            assert wr
            await wr.write_more(b'foo')
            await wr.write_more(b'bar')

        assert await _read_file(store, 'test.txt') == b'foobar'

    @pytest.mark.asyncio
    async def test_open_existing(self, store: mod.IFileStorage):
        async with store.open_file_writer('test.txt', if_exists='replace') as wr:
            assert wr
            await wr.write_more(b'foo')

        f = store.open_file_writer('test.txt')
        with pytest.raises(FileExistsError):
            async with f:
                pass
        assert await _read_file(store, 'test.txt') == b'foo'

    @pytest.mark.asyncio
    async def test_open_nonexistent(self, store: mod.IFileStorage):
        f = store.read_file("this-file-does-not-exist.txt'")
        with pytest.raises(FileNotFoundError):
            async with f:
                pass

    @pytest.mark.asyncio
    async def test_open_exist_keep(self, store: mod.IFileStorage):
        async with store.open_file_writer('test.txt', if_exists='replace') as wr:
            assert wr
            await wr.write_more(b'foo')
        async with store.open_file_writer('test.txt', if_exists='keep') as wr:
            assert wr is None
        assert await _read_file(store, 'test.txt') == b'foo'

    @pytest.mark.asyncio
    async def test_open_exist_replace(self, store: mod.IFileStorage):
        async with store.open_file_writer('test.txt', if_exists='replace') as wr:
            assert wr
            await wr.write_more(b'foo')
        async with store.open_file_writer('test.txt', if_exists='replace') as wr:
            assert wr
            await wr.write_more(b'bar')
        assert await _read_file(store, 'test.txt') == b'bar'

    @pytest.mark.asyncio
    async def test_open_multiple_simultaneous(self, store: mod.IFileStorage):
        async with store.open_file_writer('a.txt'):
            async with store.open_file_writer('b.txt'):
                pass

    @pytest.mark.asyncio
    async def test_convenience(self, store: mod.IFileStorage):
        await mod.store_file_as(__file__, 'test.py', into=store)
        content = await mod.recover('test.py', from_=store)
        assert content == Path(__file__).read_bytes()


class TestDatabaseFileStorage(CaseSet_IFileStorage):
    def do_create_storage(self, tmp_path: Path) -> AsyncContextManager[mod.IFileStorage]:
        db = Database.get_or_create(':memory:')
        return mod.open_db_storage(db=db, run_id=db.new_run_id())

    def test_create_twice(self):
        db = Database.get_or_create(':memory:')
        mod.open_db_storage(db=db, run_id=1)
        mod.open_db_storage(db=db, run_id=1)


class TestNativeFileStorage(CaseSet_IFileStorage):
    @asynccontextmanager
    async def do_create_storage(self, tmp_path: Path) -> AsyncIterator[mod.IFileStorage]:
        nfs = mod.NativeFileStorage(tmp_path)
        await fs.remove(nfs.path, recurse=True, absent_ok=True)
        yield nfs



from __future__ import annotations

import asyncio
import contextvars
import enum
from contextlib import contextmanager
from typing import Iterator, cast

from .event import ConnectionToken, Event, Handler


class CancelLevel(enum.Enum):
    """
    The level of cancellation requested by a canceller.
    """
    Request = 'req'
    'A polite cancellation request. Analogous to SIGINT.'
    Demand = 'demand'
    'A harsh cancellation request. Analagous to SIGTERM.'
    Kill = 'kill'
    'An unignorable cancellation request. Analagous to SIGKILL.'


class CancellationToken:
    """
    An object that can be used to dispatch a cancellation to an asynchronous
    operation or an operation running in another thread.
    """
    _CTX_LOCAL = contextvars.ContextVar['CancellationToken | None']('CancellationToken._CTX_LOCAL', default=None)

    def __init__(self) -> None:
        self._event = Event[CancelLevel]()
        self._cancel_level: CancelLevel | None = None

    @property
    def is_cancelled(self) -> bool:
        """Whether a cancellation has been requested"""
        return self.cancel_level is not None

    @property
    def cancel_level(self) -> CancelLevel | None:
        """The current cancellation level (or `None` if no cancellation has happened)"""
        return self._cancel_level

    def connect(self, handler: Handler[CancelLevel], *, never_immediate: bool = False) -> ConnectionToken:
        """
        Register a cancellation handler.

        :param handler: The handler for the cancellation.
        :param never_immediate: By default, if there is a pending cancellation
            request, the handler will be invoked immediately in the calling
            thread. If `never_immediate` is `True`, this immediate callback
            will not occurr. **Note:** this means a prior cancellation event
            will not be fired!
        """
        if self.is_cancelled and not never_immediate:
            assert self.cancel_level
            handler(self.cancel_level)
        return self._event.connect(handler)

    def disconnect(self, tok: ConnectionToken) -> None:
        """
        Disconnect a cancellation handler
        """
        self._event.disconnect(tok)

    def cancel(self, level: CancelLevel = CancelLevel.Request) -> None:
        """
        Issue a cancellation on this token.

        :param level: The cancellation level. Default is :attr:`~.Request`.
        """
        self._cancel_level = level
        self._event.emit(level)

    def raise_if_cancelled(self) -> None:
        """
        If a cancellation has been requested, raises
        :class:`asyncio.CancelledError`.
        """
        if self.is_cancelled:
            raise asyncio.CancelledError()

    @staticmethod
    def set_context_local(token: CancellationToken | None) -> None:
        """Set the context-local cancellation token"""
        CancellationToken._CTX_LOCAL.set(token)

    @staticmethod
    def get_context_local() -> CancellationToken | None:
        """Get the context-local cancellation token (Possibly `None`)"""
        return CancellationToken._CTX_LOCAL.get()

    @staticmethod
    @contextmanager
    def ensure_context_local() -> Iterator[CancellationToken]:
        tok = CancellationToken.get_context_local()
        if tok is None:
            tok = CancellationToken()
        with CancellationToken.scoped_context_local(tok):
            yield tok

    @classmethod
    def resolve(cls, cancel: CancellationToken | None) -> CancellationToken | None:
        if cancel is None:
            return CancellationToken.get_context_local()
        return cancel

    @classmethod
    @contextmanager
    def scoped_context_local(cls, token: CancellationToken | None) -> Iterator[CancellationToken | None]:
        """
        Create a context manager that sets the context-local cancellation token
        to `token` for the duration of the context.
        """
        prev = cls.get_context_local()
        cls.set_context_local(token)
        try:
            yield token
        finally:
            cls.set_context_local(prev)

    Shield = cast('CancellationToken', object())
    """
    :type: CancellationToken

    A cancellation token that can blocks all cancellation requests
    """


class _NeverCancelled(CancellationToken):
    def cancel(self, level: CancelLevel = CancelLevel.Request) -> None:
        pass


CancellationToken.Shield = _NeverCancelled()


def raise_if_cancelled(c: CancellationToken | None) -> None:
    """
    If given :obj:`None`, does nothing. If given a :class:`CancellationToken`,
    calls :func:`CancellationToken.raise_if_cancelled`.

    This function is used to simplify code that takes an optional
    :class:`CancellationToken` objects. Instead of testing for ``None`` at
    each cancellation point, one can simply pass the optional object to this
    function::

        async def do_stuff(seq, cancel: CancellationToken = None) -> None:
            for item in seq:
                # Create a cancellation point. No need to check for `None`:
                event.raise_if_cancelled(cancel)
                # Process an item:
                await process_item(item, cancel)
    """
    if c:
        c.raise_if_cancelled()
